#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/5/24 14:37
@File    : py_cfg.py
@Software: PyCharm
@Desc    : 
"""
import copy
import os
import pickle
import pprint
import sys
import warnings
from keyword import iskeyword
from pathlib import Path
from typing import Callable, Any, TypeVar, Type, Generator, final, cast, Iterable, Union, ClassVar, Literal

from colorama import Fore, Style

if sys.version_info >= (3, 11):
    from typing import Self, Never, TypeAlias
else:  # 兼容Python<3.11。
    Self = TypeVar('Self', bound='ADict')
    from typing import NoReturn as Never
    TypeAlias = Any

try:
    from addict import Dict
    DICT_AVAILABLE = True
except ImportError:
    Dict = dict
    DICT_AVAILABLE = False

from .open_cfg import open_config

__all__ = ['ItemLazy', 'IL', 'Config', 'ADict', 'DEP']

T = TypeVar("T")
T_dict = TypeVar('T_dict', bound=dict)
T_ADict = TypeVar('T_ADict', bound='ADict')
T_Config = TypeVar('T_Config', bound='Config')
T_IL_Func: TypeAlias = Callable[['Config'], Any]


def is_valid_variable_name(name: str) -> bool:
    return name.isidentifier() and not iskeyword(name)


def prefix2name(prefix: str) -> str:
    return prefix[:-1]  # '' -> '', 'xxx.' -> 'xxx'


def name2prefix(name: str) -> str:
    return f'{name}.' if name else name  # 'xxx' -> 'xxx.', '' -> ''


def append_prefix(prefix: str, suffix: str, still_prefix: bool=True) -> str:
    if is_valid_variable_name(suffix):
        name = f'{prefix}{suffix}'
    else:
        name = f"{prefix2name(prefix)}['{suffix}']"

    return name2prefix(name) if still_prefix else name


class ItemLazy(object):

    def __init__(self, func: T_IL_Func, priority: int=1, rel: bool=True):
        """配置惰性项。

        Args:
            func: 惰性项的计算函数。
            priority: 惰性项优先级，越小越优先。
            rel: 是否使用相对配置根。
        """
        self.func = func
        self.priority = priority
        self._level: Union[int, float, None] = None

        if not rel:  # 如果使用绝对根，则level为无穷。
            self._level = float('inf')

    @property
    def level(self) -> Union[int, float]:
        return self._level

    @level.setter
    def level(self, value: int) -> None:
        assert self._level is None
        self._level = value

    def __call__(self, cfg: 'Config') -> Any:
        return self.func(cfg)

    @classmethod
    def ordered_item_lazy(cls, config: T_Config) -> list[tuple[T_Config, Any, int]]:
        return sorted(((c, k, l.priority) for c, k, l in config.ckl if isinstance(l, cls)), key=lambda x: x[2])

    @classmethod
    def compute_item_lazy(cls, config: T_Config) -> T_Config:
        item_lazy_dic_keys = cls.ordered_item_lazy(config)

        if len(item_lazy_dic_keys) > 0:
            for item_lazy_dic, item_lazy_key, _ in item_lazy_dic_keys:
                item_lazy: cls = item_lazy_dic[item_lazy_key]
                root, _ = item_lazy_dic.find_root(item_lazy.level)
                item_lazy_dic[item_lazy_key] = item_lazy(root)
            return cls.compute_item_lazy(config)
        else:
            return config

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.func}, priority={self.priority}, level={self.level})"


DEP = IL = ItemLazy


class ADict(Dict):
    """继承自addict.Dict，用作数据容器。修复了addict.Dict的诸多问题。"""

    _go_in_list: ClassVar[bool] = True  # 构造时递归进入list。

    # -* 构造与反构造。

    def __init__(self, *args, **kwargs):
        dict.__init__(self)

        # -* 预设__parent、__key、__frozen、_whole。
        self.set_attribute('__parent', kwargs.pop('__parent', None))
        self.set_attribute('__key', kwargs.pop('__key', None))
        self.set_attribute('__frozen', False)
        self.set_attribute('_whole', False)

        # -* CHANGE 与标准字典一样，只允许一个arg。
        if len(args) > 1:
            raise TypeError(f"{self.__class__} expected at most 1 arguments, got {len(args)}")
        elif len(args) == 1:
            arg = args[0]
            if isinstance(arg, dict):
                dic = arg
            else:
                dic = dict(arg)
            self.from_dict(dic, ret=self, memo=None, always_copy=True, deepcopy=False)
        else:
            pass

        # CHANGE kwargs不再硬update，而是先变类拷贝为ADict，再update到当前ADict上。
        if kwargs:
            self.dict_update(self.from_dict(kwargs, ret=None, memo=None, always_copy=True, deepcopy=False))

    @staticmethod
    def is_subtree(tree: Any, root: dict) -> bool:
        """判断tree是否是root的子树。"""
        return isinstance(tree, dict)

    def copy_cfg_attr_to(self, other: 'ADict') -> None:
        if isinstance(other, ADict):
            other.set_attribute('_whole', self.get_attribute('_whole'))

    @classmethod
    def from_dict(cls: Type[T_ADict],
                  other: dict, ret: T_ADict=None, memo: dict=None,
                  always_copy: bool=False, deepcopy: bool=False) -> T_ADict:
        # -* 参数检查。
        if not isinstance(other, dict):
            raise TypeError(f"{cls.__name__}.from_dict expected dict as other, got {type(other)}")

        if ret is not None:  # ret不为None只用于初始化的时候，此时总是做浅拷贝，且因为是从头开始，memo应当为None。
            assert always_copy and (memo is None) and (not deepcopy), "Param 'ret' can only be used by __init__."

        # -* 参数预处理。
        if memo is None:
            memo = {}

        if deepcopy:  # deepcopy时，总是拷贝。
            always_copy = True

        # -* 若不是always_copy，说明目的只是变类。若other是cls的实例，则可以不做拷贝，直接返回。
        if (not always_copy) and (type(other) is cls):  # 此时deepcopy=False, ret=None。
            assert len(memo) == 0  # 若memo不为空，要么是处于deepcopy状态，要么是处于_from_dict_hook后的栈，都不是变类用法。
            return cast(T_ADict, other)

        # -* 开启拷贝。
        if id(other) in memo:  # 若other已经被拷贝过，则直接返回拷贝。
            return memo[id(other)]

        if ret is None:  # 构造新树。
            ret = cls()
        else:  # 新树由__init__提供。
            assert type(ret) is cls and len(ret) == 0, f"{ret=} should be an empty {cls=}."
        memo[id(other)] = ret  # 将新容器加入memo，处理循环引用。

        if isinstance(other, ADict):  # 如果是ADict及其子类，则也要做属性拷贝。
            other.copy_cfg_attr_to(ret)

        for k, v in other.items():
            ret[k] = cls._from_dict_hook(v, other, memo=memo, deepcopy=deepcopy)

        return ret

    @classmethod
    def _from_dict_hook(cls, item: Any, other: dict, memo: dict, deepcopy: bool) -> Any:
        if id(item) in memo:  # 若item已经被拷贝过，则直接返回拷贝。
            ret = memo[id(item)]
        elif cls.is_subtree(item, other):
            ret = cls.from_dict(item, ret=None, memo=memo, always_copy=True, deepcopy=deepcopy)
        elif cls._go_in_list and isinstance(item, list):
            ret = []
            memo[id(item)] = ret
            ret.extend(cls._from_dict_hook(elem, other, memo=memo, deepcopy=deepcopy) for elem in item)
        else:
            ret = copy.deepcopy(item, memo) if deepcopy else item

        return ret

    def to_dict(self, memo: dict=None) -> dict:
        # -* 参数预处理。
        if memo is None:
            memo = {}

        # -* 开启拷贝。
        if id(self) in memo:  # 若other已经被拷贝过，则直接返回拷贝。
            return memo[id(self)]

        ret = {}  # 构造新树。

        memo[id(self)] = ret  # 将新容器加入memo，防止循环引用。

        for k, v in self.items():
            ret[k] = self._to_dict_hook(v, memo=memo)
        return ret

    @classmethod
    def _to_dict_hook(cls, item: Any, memo: dict) -> Any:
        if id(item) in memo:  # 若item已经被拷贝过，则直接返回拷贝。
            ret = memo[id(item)]
        elif isinstance(item, ADict):
            ret = item.to_dict(memo=memo)
        elif DICT_AVAILABLE and isinstance(item, Dict):
            ret = item.to_dict()  # 兼容addict.Dict，但memo无法传递。
        elif cls._go_in_list and isinstance(item, list):  # 保持对称，若from_dict时list被视为普通值，则to_dict时也应该如此。
            ret = []
            memo[id(item)] = ret
            ret.extend(cls._to_dict_hook(elem, memo=memo) for elem in item)
        else:
            ret = item

        return ret

    # -* 序列化与反序列化。

    def __getnewargs__(self) -> tuple:  # CHANGE 无需返回任何值。
        return tuple()

    def __getstate__(self) -> Any:  # CHANGE 返回self.__dict__，而非自身作为state。
        if sys.version_info >= (3, 11):
            return object.__getstate__(self)
        else:
            return vars(self)

    def __setstate__(self, state: dict) -> None:  # CHANGE: 恢复object默认的__setstate__。
        vars(self).update(state)  # 用户自定义的类，哪怕没有__init__过，其实例总是有__dict__。

    # -* 属性操作。

    def get_attribute(self, name: str, /) -> Any:
        return object.__getattribute__(self, name)

    def get_attribute_default(self, name: str, default: Any=None, /) -> Any:
        try:
            return self.get_attribute(name)
        except AttributeError:
            return default

    def __getattr__(self, item: str, /) -> Any:
        return self[item]

    def __setattr__(self, name: str, value: Any, /) -> None:
        if hasattr(self.__class__, name):
            raise AttributeError(f"{self.__class__}'s class attribute '{name}' is read-only.")
        elif name in vars(self):  # CHANGE 实例属性也不可冲突。
            raise AttributeError(f"{self.__class__}'s instance attribute '{name}' is read-only.")
        else:
            self[name] = value

    def __delattr__(self, name, /) -> None:
        del self[name]

    def set_attribute(self, name: str, value: Any, /) -> None:
        object.__setattr__(self, name, value)

    def del_attribute(self, name: str, /) -> None:
        object.__delattr__(self, name)

    # -* 字典操作。

    def _mount2parent(self) -> None:
        if (p := self.get_attribute('__parent')) is not None:
            p[self.get_attribute('__key')] = self

            self.set_attribute('__parent', None)  # CHANGE pk不删除，而是置为None。
            self.set_attribute('__key', None)

    def __setitem__(self, key: Any, value: Any, /) -> None:
        if not vars(self):  # CHANGE 若实例属性字典为空，说明处于pickle.loads的恢复内生数据阶段，可走捷径直接dict setitem。
            dict.__setitem__(self, key, value)
            return

        self._setitem(key, value)

    def _setitem(self, key: Any, value: Any, /) -> None:
        if self.is_frozen:  # CHANGE 调用addict setitem前，检查frozen，若是，则完全禁止setitem。
            raise RuntimeError(f"{self.__class__} is frozen. ")

        self._mount2parent()

        dict.__setitem__(self, key, value)

    def __missing__(self, key: Any, /) -> Self:
        if self.is_frozen:
            raise KeyError(f"{key} not found in frozen {self.__class__}.")
        return self.__class__(__parent=self, __key=key)

    def __delitem__(self, key, /) -> None:
        if self.is_frozen:
            raise RuntimeError(f"{self.__class__} is frozen. ")

        dict.__delitem__(self, key)

    # -* 拷贝与更新。

    def branch_copy(self) -> Self:
        return self.from_dict(self, ret=None, memo=None, always_copy=True, deepcopy=False)

    def copy(self) -> Self:
        return self.branch_copy()

    def __copy__(self) -> Self:
        return self.branch_copy()

    def deepcopy(self) -> Self:
        return self.from_dict(self, ret=None, memo=None, always_copy=True, deepcopy=True)

    def __deepcopy__(self, memo: dict) -> Self:
        return self.from_dict(self, ret=None, memo=memo, always_copy=True, deepcopy=True)

    @classmethod
    def _dict_update(cls, dic: dict, other: dict, incremental: bool=False, copy_other: bool=True) -> None:
        # -* 参数检查。
        if not isinstance(dic, dict):
            raise TypeError(f"{cls.__name__}._dict_update expected dict as dic, got {type(dic)}")

        if not isinstance(other, dict):
            raise TypeError(f"{cls.__name__}._dict_update expected dict as other, got {type(other)}")

        # NOTE 可以证明，对ADict的子树判定（dict子类-dict子类），拷贝前互为子树 ⇔ 拷贝后互为子树，且均为ADict类。
        # NOTE 可以证明，对Config的子树判定（D-D），拷贝前互为子树 ⇒ 拷贝后互为子树，且均为Config类，
        # NOTE 但是 拷贝后互为子树，且均为Config类 不能推出 拷贝前互为子树。
        # NOTE 总而言之，可以认为拷贝能保留子树关系，且子树有is_whole性质。但Config的拷贝可能会建立新的子树关系。
        other: ADict = (cls.from_dict(other, ret=None, memo=None, always_copy=True, deepcopy=False)
                        if copy_other else
                        other)

        # NOTE dict_update只更新k-v对，不涉及更新属性（依赖、是否不可分）。
        # NOTE 因为除了以下场景，想不到需要更新属性的情况：
        # NOTE 1）更新依赖，但这种情况下，应当用_whole完整覆盖原来的树。本打算继承的部分可以独立出来，作为共享的导入。
        for k, v in tuple(other.items()):
            # NOTE 每对键值都只使用一次，故取出后就可以从other中删除，可以防止Config一颗子树两个爹。
            # NOTE 该做法的风险只发生在other中存在循环引用。目前循环引用下的行为是未定义的，干脆不考虑。
            del other[k]

            # 若人有我有，且都是子树，且人的子树不被视作整体，则子树更新子树。
            # NOTE 若某棵树_whole=True，改树作为value时，视作不可递归的整体。
            if ((k in dic) and
                    cls.is_subtree(v, other) and (not isinstance(v, ADict) or not v.is_whole) and
                    cls.is_subtree(dic[k], dic)):
                cls._dict_update(dic[k], v, incremental=incremental,
                                 copy_other=False)  # 只有根的other需要拷贝，后续递归都是直接更新。
            else:
                if incremental:  # 若增量式，则人有我无，方才更新。
                    if k not in dic:
                        dic[k] = v
                    else:
                        pass
                else:  # 人有我有，但至少一方不是子树；或人有我无，则直接更新。
                    dic[k] = v

    def dict_update(self, other: dict, incremental: bool=False) -> None:
        """接收一个dict类型（或其子类）的配置树，更新到当前配置树上。

        应当确保，此后对本配置树枝干的修改，不会同步到原配置树上。因此，应当拷/转枝赋叶原配置树后，再做更新。

        Args:
            other: dict类型的配置树。
            incremental: 增量式更新，如果冲突，保留旧树。

        Returns:
            更新后的配置树。
        """
        self._dict_update(self, other, incremental=incremental)

    def update(self, *args, **kwargs) -> None:
        if len(args) > 1:
            raise TypeError(f"{self.__class__}.update expected at most 1 arguments, got {len(args)}")
        elif len(args) == 1:
            self.dict_update(args[0])
        else:
            pass

        if kwargs:
            self.dict_update(kwargs)

    # -* 运算。
    def __add__(self, other: Never) -> Never:
        raise NotImplementedError("ADict does not support __add__.")

    def __or__(self, other: dict) -> Self:
        if not isinstance(other, dict):
            return NotImplemented(f"{self.__class__} does not support | with {type(other)}.")
        new = self.branch_copy()
        new.dict_update(other)
        return new

    def __ror__(self, other: dict) -> Self:
        return self.__or__(other)

    def __ior__(self, other: dict) -> Self:
        if not isinstance(other, dict):
            return NotImplemented(f"{self.__class__} does not support |= with {type(other)}.")
        self.dict_update(other)
        return self

    # -* 冻结操作。

    def freeze(self, shouldFreeze=True) -> Self:
        for b in self.branches:
            if isinstance(b, Dict if DICT_AVAILABLE else ADict):  # 将所有子树也冻结。
                object.__setattr__(b, '__frozen', shouldFreeze)
        return self

    def unfreeze(self) -> Self:
        return self.freeze(False)

    @property
    def is_frozen(self) -> bool:
        return self.get_attribute_default('__frozen', False)

    # -* 重写受到影响的字典方法。

    def setdefault(self, key: Any, default: Any=None) -> Any:
        if key in self:
            return self[key]
        else:
            self[key] = default
            return default

    @classmethod
    def fromkeys(cls: Type[T_ADict], iterable: Iterable[T], value=None) -> T_ADict:
        return cls.from_dict(dict.fromkeys(iterable, value), ret=None, memo=None, always_copy=True, deepcopy=False)

    def clear(self) -> None:
        if self.is_frozen:
            raise RuntimeError(f"{self.__class__} is frozen. ")

        dict.clear(self)

    def pop(self, key: Any, default: Any=None) -> Any:
        if self.is_frozen:
            raise RuntimeError(f"{self.__class__} is frozen. ")

        return dict.pop(self, key, default)

    def popitem(self) -> tuple[Any, Any]:
        if self.is_frozen:
            raise RuntimeError(f"{self.__class__} is frozen. ")

        return dict.popitem(self)

    # -* _whole属性相关操作。

    def set_whole(self, is_whole: bool=True) -> Self:
        """将自身设置为不可递归的整体。"""
        self.set_attribute('_whole', is_whole)
        self._mount2parent()  # 挂载到父节点，确保设置总是有效。
        return self

    override = set_whole

    @property
    def is_whole(self) -> bool:
        return self.get_attribute_default('_whole', False)

    def empty_leaf(self) -> Self:
        """转为本类的空且whole的字典，即空叶子。常用于表示无效/无定义项。"""
        self.clear()
        return self.set_whole(True)  # 设置为整体，并挂载到父节点。

    def empty_tree(self) -> Self:
        """转为本类的空的字典，即空子树。"""
        self.clear()
        return self.set_whole(False)

    # -* 浏览。

    @classmethod
    def _named_branches(cls, dic: dict, name: str='',
                        memo: set=None) -> Generator[tuple[str, dict], None, None]:
        if memo is None:
            memo = set()

        if id(dic) in memo:  # 若已经遍历过，截断。
            return
        memo.add(id(dic))  # 该树已经遍历过。

        yield name, dic

        for k, v in dic.items():
            if cls.is_subtree(v, dic):
                subtree_name = append_prefix(name2prefix(name), k, still_prefix=False)
                yield from cls._named_branches(v, subtree_name, memo)  # 若v是子树，则抛出子树的枝条。

    @property
    def named_branches(self) -> Generator[tuple[str, dict], None, None]:
        yield from self._named_branches(self)

    @property
    def branches(self) -> Generator[dict, None, None]:
        yield from (b for _, b in self._named_branches(self))

    @classmethod
    def _named_ckl(cls, dic: dict, prefix: str= '',
                   memo: set=None) -> Generator[tuple[str, tuple[dict, Any, Any]], None, None]:
        if memo is None:
            memo = set()

        if id(dic) in memo:  # 若已经遍历过，截断。
            return
        memo.add(id(dic))  # 该树已经遍历过。

        for k, v in dic.items():
            if cls.is_subtree(v, dic):
                pre = append_prefix(prefix, k, still_prefix=True)
                yield from cls._named_ckl(v, pre, memo)  # 若v是子树，则抛出子树的named_ckl。
            else:
                name = append_prefix(prefix, k, still_prefix=False)
                yield name, (dic, k, v)  # 若v是叶子，则抛出名字、父节点、键、值。

    @property
    def named_ckl(self) -> Generator[tuple[str, tuple[dict, Any, Any]], None, None]:
        yield from self._named_ckl(self)

    @property
    def ckl(self) -> Generator[tuple[dict, Any, Any], None, None]:
        yield from ((c, k, l) for _, (c, k, l) in self.named_ckl)

    @property
    def leaves(self) -> Generator[Any, None, None]:
        yield from (l for _, (_, _, l) in self.named_ckl)

    # -* 快捷方式。

    def set_func(self, name: Union[str, None]=None) -> Callable[[Callable], Callable]:
        """返回装饰器，装饰器将被装饰函数注册为当前配置树的项目。

        Args:
            name: 函数名，若为None，则使用被装饰函数的函数名。

        Returns:
            装饰器。
        """

        def decorator(func):
            assert callable(func), f"IL must be callable, but got {func}"

            if name is not None:
                self[name] = func
            else:
                self[func.__name__] = func

            return func

        return decorator

    # -* 打印。

    def to_txt(self, prefix: str='dic.', print_empty_leaf: bool=False, color: bool=False) -> str:
        def cr(n: str, fore: str) -> str:
            if color:
                return f'{fore}{n}{Style.RESET_ALL}'
            else:
                return n

        if len(prefix) == 0:  # 为了可读性，禁止空前缀。
            raise ValueError("prefix should not be empty.")

        if len(prefix) > 0 and not prefix.endswith('.'):  # 前缀应当以'.'结尾。
            prefix = name2prefix(prefix)

        root_name = prefix2name(prefix)
        lines = [f'{cr(root_name, Fore.MAGENTA)} = {self.__class__.__name__}()']

        # NOTE 特殊树在此处构造。
        for name, b in self._named_branches(self, name=root_name):
            # 若为_whole或有caps，则为ADict特殊树。特殊树要打印特殊属性。
            if ((isinstance(b, ADict) and b.is_whole) or
                    (isinstance(b, Config) and b.get_attribute('_cfgs_update_at_parser'))):
                line = f"{cr(name, Fore.MAGENTA)}.override({b.is_whole})"  # 打印ADict的_whole属性，该操作还会确保树的存在。
                # 若为Config且存在依赖，还要打印_cfgs_update_at_parser。
                if isinstance(b, Config) and (caps := b.get_attribute('_cfgs_update_at_parser')):
                    line = line + f".set_attribute('_cfgs_update_at_parser', {caps})"
                lines.append(line)

        # NOTE 所有值和普通非空树在此处构造。
        lines.append('# ------- ↓ LEAVES ↓ ------- #')

        for name, (c, k, l) in self._named_ckl(self, prefix=prefix):
            str_l = pprint.pformat(l)
            if '\n' not in str_l:
                lines.append(f"{cr(name, Fore.CYAN)} = {str_l}")
            else:
                lines.append(f"{cr(name, Fore.CYAN)} = \\ \n{str_l}")

        # NOTE 打印空树。
        if print_empty_leaf:
            lines.append('# ------- ↓ EMPTY LEAVES ↓ ------- #')

            for name, b in self._named_branches(self, name=root_name):
                if (len(b) == 0) and (not (isinstance(b, ADict) and b.is_whole)):
                    lines.append(f"{cr(name, Fore.YELLOW)} = {b.__class__.__name__}()")

        return '\n'.join(lines)


class Config(ADict):
    """配置字典。继承自Dict，从类型上和Addict区分，以便分离配置项和配置树。"""

    _go_in_list: ClassVar[bool] = False  # 构造时不递归进入list。

    # -* Config的初始化与解析。

    def __init__(self, *cfgs: Union[dict, str],
                 cfgs_update_at_parser: Union[tuple[str, ...], str]=(), caps: Union[tuple[str, ...], str]=(), **kwargs):
        """支持从其他其他配置树模块路径或配置树dict初始化。所有配置树会被逐个dict_update到当前配置树上。

        Args:
            *cfgs: List[配置树所在模块|配置树]
            cfgs_update_at_parser: 解析时用于更新的基配置。
            caps: cfgs_update_at_parser的别名。
            **kwargs: 传递给Dict，不应该使用。
        """
        # -* 初始化。
        __mounted = kwargs.pop('__mounted', True)
        super().__init__(**kwargs)  # __parent, __key, __frozen, _whole被设置。
        self.set_attribute('__mounted', __mounted)

        # -* 检查caps。
        if caps:
            assert not cfgs_update_at_parser, f"{caps=}和{cfgs_update_at_parser=}不能同时使用。"
            cfgs_update_at_parser = caps

        # -* 同一cfgs_update_at_parser格式。
        if isinstance(cfgs_update_at_parser, str):
            cfgs_update_at_parser = (cfgs_update_at_parser,)

        # -* 遍历配置树，更新到当前配置树上。
        for i, cfg in enumerate(cfgs):
            # * 若给出配置树模块路径，则根据路径打开配置树。
            if isinstance(cfg, str):
                cfg, _ = open_config(cfg)
            # * 检查配置树是dict。
            if not isinstance(cfg, dict):
                raise ValueError(f"{i}'th item = {cfg} in cfgs should be (opened as) dict.")
            # * 将配置树更新到本配置树上。
            self.dict_update(cfg)
            # * 将配置树的parser时配置更新上来，且优先级更高。
            if isinstance(cfg, Config) and (cfg_dep := cfg.get_attribute('_cfgs_update_at_parser')):
                # A --解析时--> B; D, B --解析时--> C 含义明确，即加载时DFS，从祖先开始增量更新到C，优先级为B、A、D。
                # A --解析时--> B; D, B --加载时--> C 若D与A冲突，则D中键值会阻塞A，优先级为B、D、A，这是我们不希望的。
                # 因此，最好不要混用解析时和加载时依赖。
                warnings.warn(f"{cfg=}存在解析时依赖{cfg_dep=}。\n"
                              f"因此，该配置也应该作为当前配置的解析时依赖。否则并列的加载时依赖，可能干扰cfg_dep更新cfg。")
                cfgs_update_at_parser = cfgs_update_at_parser + cfg_dep

        # -* 记录基配置。
        # self.set_attribute('_cfgs_update_at_init', cfgs)  # ! 该值除了debug，根本用不到。如果cfgs不是str而是字典，pickle还会增加额外的负担。
        self.set_attribute('_cfgs_update_at_parser', cfgs_update_at_parser)

    def update_at_parser(self) -> None:
        # -* 获取解析式基配置。
        cfgs_update_at_parser = self.get_attribute('_cfgs_update_at_parser')
        # -* 若无需更新，跳过。
        if not cfgs_update_at_parser:
            return
        # -* 保存主配置。
        main_cfg = self.branch_copy()
        # -* 清空当前配置树。
        self.clear()
        # -* 逐个读取基配置，并更新到当前配置中。
        for i, base_cfg in enumerate(cfgs_update_at_parser):
            # -* 若给出配置树模块路径，则根据路径打开配置树。
            if isinstance(base_cfg, str):
                base_cfg, _ = open_config(base_cfg)
            # -* 基配置也进行解析时更新。
            if isinstance(base_cfg, Config):
                base_cfg.update_at_parser()
            # -* 检查配置树是dict。
            if not isinstance(base_cfg, dict):
                raise ValueError(f"{i}'th item = {base_cfg} in self._cfgs_update_at_parser should be "
                                 f"(opened as) dict.")
            # -* 将基配置树更新到本配置树上。
            self.dict_update(base_cfg)
        # * 将主配置树更新回当前配置树。
        self.dict_update(main_cfg)

    def parse(self, experiments_root: str='', config_root: str= './configs',
              create_rslt_dir: bool=True) -> Self:
        # TODO 避免循环导入。Ugly，更好的办法是将parse_cfg.py中的parse_config放到py_cfg.py中。
        from .parse_cfg import parse_config
        return parse_config(self, experiments_root, config_root, create_rslt_dir)

    def compute_item_lazy(self) -> Self:
        return ItemLazy.compute_item_lazy(self)

    def load(self, experiments_root: str='', config_root: str='./configs',
             create_rslt_dir: bool=True) -> 'Config':
        # TODO 避免循环导入。Ugly，更好的办法是将parse_cfg.py中的load_config放到py_cfg.py中。
        from .parse_cfg import load_config
        return load_config(self, experiments_root, config_root, create_rslt_dir)

    @property
    def subtrees_wt_COM(self: T_ADict) -> list[T_ADict]:
        return [subtree for subtree in self.branches if 'COM_' in subtree]

    def _check_COM(self: T_ADict):
        """检查COM树是否合规。由于reduce_COM会逐步删除子COM树，可能影响安全检查。故检查放在reduce_COM之前。"""
        warnings.warn("COM feature is deprecated", DeprecationWarning)
        # NOTE COM树，可以有子COM树。因为子树搜索为DFS后序，故子树COM树总是先于父COM树处理。
        # NOTE 而COM并列项如果有子COM树，则无法确定和当前COM项的执行顺序，非常危险，要警告。
        for subtree_wt_COM in self.subtrees_wt_COM:
            # * COM项应该是子树。
            if not self.is_subtree(subtree_wt_COM.COM_, self):
                raise RuntimeError(f"COM树{subtree_wt_COM}的COM项{subtree_wt_COM.COM_}应当为子树。")
            # * COM并列项不含COM子树。
            COM_parallel_vals = [v for k, v in subtree_wt_COM.items() if (k != 'COM_') and (self.is_subtree(v, self))]
            for par_val in COM_parallel_vals:
                par_val: T_ADict
                if par_val.subtrees_wt_COM:
                    warnings.warn(f"COM树{subtree_wt_COM}的COM并列项{par_val}存在子COM树。无法确定该树和当前"
                                  f"COM树的执行顺序，reduce_COM行为不可预测。")

    def reduce_COM(self: T_ADict):
        """归并掉所有COM项。"""
        warnings.warn("COM feature is deprecated", DeprecationWarning)
        self._check_COM()  # 合规检查。
        for subtree_wt_COM in self.subtrees_wt_COM:
            # * 获得COM项和COM并列项。
            COM_val = subtree_wt_COM.COM_
            COM_parallel_vals = [v for k, v in subtree_wt_COM.items() if (k != 'COM_') and (self.is_subtree(v, self))]
            # * 将COM项增量更新到并列项上。
            for par_val in COM_parallel_vals:
                par_val: T_ADict
                # NOTE 当_COM树本身为_whole时：1）cfg_update时，作为不可递归树覆盖；2）更新并列树时，行为不变（此时_COM树
                # NOTE 作为根节点树，_whole无效。
                # NOTE 当_COM树的子树为_whole时：1）cfg_update时，作为不可递归树覆盖；2）更新并列树时, 作为不可递归树覆盖。
                # NOTE 该行为可能与预期不符（我们希望更新并列树时，不要覆盖），此时应当寻找其他workaround，如让_COM树，而非
                # NOTE _COM树的子树为_whole。
                par_val.dict_update(COM_val, incremental=True)
            # * 删除COM项。
            del subtree_wt_COM['COM_']

    # -* 重载子树定义和属性拷贝。

    @staticmethod
    def is_subtree(tree: Any, root: dict) -> bool:
        """判断tree是否是root的子树。"""
        return type(tree) is type(root)

    @final  # 不应该重载该方法。所有Config子类应当由相同的copy_cfg_attr_to，否则from_dict无法在不同Config子类间正确运行。
    def copy_cfg_attr_to(self, other: ADict) -> None:
        """将self的cfg属性(依赖、是否不可分)拷贝到other上。"""
        super().copy_cfg_attr_to(other)  # 拷贝_whole属性。
        if isinstance(other, Config):
            # other.set_attribute('_cfgs_update_at_init', self.get_attribute('_cfgs_update_at_init'))
            other.set_attribute('_cfgs_update_at_parser', self.get_attribute('_cfgs_update_at_parser'))

    # -* 双向树结构。

    @property
    def _is_mounted(self) -> bool:
        # 若没有_mounted，则处于pickle的new而未init阶段，此时无需担心挂载问题。
        return self.get_attribute_default('__mounted', True)

    def _mount2parent(self) -> None:
        if not self._is_mounted:
            self.get_attribute('__parent')[self.get_attribute('__key')] = self
            self.set_attribute('__mounted', True)

    def demount(self) -> None:
        assert self._is_mounted
        self.set_attribute('__parent', None)
        self.set_attribute('__key', None)

    def find_root(self, level: Union[int, float]=float('inf')) -> tuple[Self, int]:
        """寻找树根。"""
        root = self
        count = level
        cur_level = 0

        while count:
            new_root = root.get_attribute('__parent')
            if new_root is None:
                if count == float('inf'):  # level为inf时，穷尽所有父节点寻找根。
                    break
                else:
                    raise RuntimeError(f"Can only find root at max level={cur_level}, but got level={level}.")
            else:
                root = new_root

            count -= 1
            cur_level += 1

        return root, cur_level

    def _setitem(self, key: Any, value: Any, /) -> None:
        super()._setitem(key, value)  # frozen检查、挂载、dict.__setitem__。
        value = self[key]  # 重新获取value，不能保证value在设置后不变。

        # -* 确保子树自洽。
        if self.is_subtree(value, self):
            value: Config
            # -* 对missing情况，检查parent、key正确，防止一子二父。
            if not value._is_mounted:
                assert value.get_attribute('__parent') is self
                assert value.get_attribute('__key') == key  # key是不可变对象，故用==比较。
            # -* 非missing情况，对已经有parent的子树，替换为它的拷贝，防止一子二父。
            elif (value.get_attribute('__parent') is not None) or (value.get_attribute('__key') is not None):
                self[key] = value.branch_copy()
                value = self[key]
                # TODO 此处有一个隐藏BUG：当该value含有DEP时，改DEP会被赋值过来。若该DEP是rel模式，则level很可能错乱。
                # TODO 解决方案1：重算level和lambda：太复杂，做不到。
                # TODO 解决方案2：拷贝过来的IL作为“分身”，不直接计算，等“真身”计算完后赋值：“真身”若来自其他树，可能不计算，不可信。
                # TODO 解决方案3：安全检查，赋值过来的树里，如果有rel的DEP，如果其level超过了赋值树的顶层，则报错。
            # -* 根子树挂载。
            else:
                value.set_attribute('__parent', self)
                value.set_attribute('__key', key)

        # -* 对IL，设置其层级。
        if isinstance(value, ItemLazy) and value.level is None:
            # 如果IL已经有level，要么是inf，说明是绝对根，不用记录level。要么是已经被设置，当前是branch_copy之类的，反正不用再设置。
            _, value.level = self.find_root(level=float('inf'))

    def __missing__(self, key: Any, /) -> Self:  # CHANGE frozen也返回空字典，反正setitem一定会触发报错。
        return self.__class__(__parent=self, __key=key, __mounted=False)

    def __delitem__(self, key, /) -> None:
        if self.is_frozen:
            raise RuntimeError(f"{self.__class__} is frozen. ")

        if self.is_subtree(v := self[key], self):
            v.demount()
        dict.__delitem__(self, key)

    def clear(self) -> None:
        if self.is_frozen:
            raise RuntimeError(f"{self.__class__} is frozen. ")

        for v in self.values():
            if self.is_subtree(v, self):
                v.demount()
        dict.clear(self)

    def pop(self, key: Any, default: Any=None) -> Any:
        if self.is_frozen:
            raise RuntimeError(f"{self.__class__} is frozen. ")

        if key not in self:
            return default

        ret = dict.pop(self, key)
        if self.is_subtree(ret, self):
            ret.demount()

        return ret

    def popitem(self) -> tuple[Any, Any]:
        if self.is_frozen:
            raise RuntimeError(f"{self.__class__} is frozen. ")

        k, v = dict.popitem(self)
        if self.is_subtree(v, self):
            v.demount()
        return k, v

    # -* 快捷方式。

    def set_IL(self, name: Union[str, None]=None, priority: int=1, rel: bool = True) \
            -> Callable[[T_IL_Func], T_IL_Func]:
        """返回装饰器，装饰器将被装饰函数注册为当前配置树的惰性项。

        Args:
            name: 函数名，若为None，则使用被装饰函数的函数名。
            priority: 惰性项的优先级。
            rel: 是否是相对优先级。

        Returns:
            装饰器。
        """

        def decorator(func):
            assert callable(func), f"IL must be callable, but got {func}"

            if name is not None:
                self[name] = IL(func, priority=priority, rel=rel)
            else:
                self[func.__name__] = IL(func, priority=priority, rel=rel)

            return func

        return decorator

    set_DEP = set_IL

    # -* 对Config，使用to_txt字符化。

    def to_txt(self, prefix: str='cfg.', print_empty_leaf: bool=False, color: bool=False) -> str:
        return super().to_txt(prefix=prefix, print_empty_leaf=print_empty_leaf, color=color)

    def __str__(self) -> str:
        return self.to_txt(color=True)

    def save_py(self, file: Union[str, os.PathLike]) -> Path:
        (file := Path(file)).parent.mkdir(parents=True, exist_ok=True)

        cfg_str = f"""
# -*- THIS CONFIG FILE IS AUTO-GENERATED BY AlchemyCat -*-
from alchemy_cat.dl_config import {type(self).__name__}

{self.to_txt(prefix='cfg.', print_empty_leaf=True)}
"""

        file.write_text(cfg_str)

        return file

    def save_pkl(self, file: Union[str, os.PathLike], save_copy: bool=True) -> Path:
        (file := Path(file)).parent.mkdir(parents=True, exist_ok=True)

        file.write_bytes(pickle.dumps(self.branch_copy() if save_copy else self))

        return file

    # -* 迁移。

    def _save_cfg_for_other_cfg_system(self) -> dict:
        # -* 分离。
        saved_cfg = self.branch_copy()

        # -* 假设该配置拥有rslt_dir，则会经历：update_at_parser -> compute_item_lazy -> freeze
        saved_cfg = saved_cfg.load(experiments_root='', create_rslt_dir=False)

        # -* 转为普通dict返回。
        return saved_cfg.to_dict()

    def save_yaml(self, file: Union[str, os.PathLike]) -> Path:
        import yaml

        (file := Path(file)).parent.mkdir(parents=True, exist_ok=True)

        file.write_text(yaml.dump(self._save_cfg_for_other_cfg_system()))

        return file

    def save_mmcv(self, file: Union[str, os.PathLike]) -> Path:
        from mmengine import Config as MMConfig

        (file := Path(file)).parent.mkdir(parents=True, exist_ok=True)

        # AlchemyCat的`rslt_dir`等价于mmcv的`work_dir`。若AlchemyCat有此意，而mmcv未定义，则补充定义。
        if ('rslt_dir' in self) and ('work_dir' not in self):
            self['work_dir'] = self['rslt_dir']

        mm_config = MMConfig(self._save_cfg_for_other_cfg_system())
        mm_config.dump(file)

        return file

    @classmethod
    def from_x_to_y(cls: type[T_Config], x: Union[str, os.PathLike], y: Union[str, os.PathLike],
                    y_type: Literal['yaml', 'mmcv', 'alchemy-cat']='alchemy-cat') -> T_Config:
        """Convert a Config, yaml, yacs or MMCV-Config file to a desired file type.

        Args:
            x: A Config, yaml, yacs or MMCV-Config file.
            y: The target file path.
            y_type: The target file type. Choices are ['yaml', 'mmcv', 'alchemy-cat'].

        Returns:
            `Config` object.
        """
        # -* 读取dict_cfg。
        dict_cfg, _ = open_config(str(x))
        # -* 转为Config。
        cfg = cls.from_dict(dict_cfg, ret=None, memo=None, always_copy=True, deepcopy=False)
        # -* 保存。
        if y_type == 'yaml':
            cfg.save_yaml(y)
        elif y_type == 'mmcv':
            cfg.save_mmcv(y)
        elif y_type == 'alchemy-cat':
            cfg.save_py(y)
        else:
            raise ValueError(f"{y_type=} is not supported.")

        return cfg

    def merge_from_file(self, cfg_filename: Union[str, os.PathLike]):  # yacs兼容。
        self.unfreeze()
        self.dict_update(open_config(str(cfg_filename))[0])
        self.freeze()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/6/26 20:19
@File    : 生成DummyTest.py
@Software: PyCharm
@Desc    : 
"""
import os
import os.path as osp

from alchemy_cat.contrib.voc import VOCAug2
from alchemy_cat.data.plugins import arr2PIL
from tqdm import tqdm

os.makedirs(dummy_dir := 'temp/DummyTest', exist_ok=True)

dt = VOCAug2('datasets', split='test')

for inp in tqdm(dt, desc="生成DummyTest", unit="张", dynamic_ncols=True):
    img_id, dummy_lb = inp.img_id, inp.lb
    arr2PIL(dummy_lb).save(osp.join(dummy_dir, f'{img_id}.png'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/7/31 17:00
@File    : auto_allocate_cuda.py
@Software: PyCharm
@Desc    : 自动分配cuda设备。
"""
import os

from .get_cuda import block_get_available_cuda, get_cudas, cudas2CUDA_VISIBLE_DEVICES

__all__ = ['allocate_cuda_by_group_rank']


def allocate_cuda_by_group_rank(group_rank: int,
                                group_cuda_num: int=None, group_num: int=None,
                                block: bool=True, verbosity: bool=True,
                                memory_need: float=-1., max_process: int=-1) -> tuple[list[int], dict[str, str]]:
    """Auto allocate cuda.

    Args:
        group_rank: Task group index.
        group_cuda_num: Number of devices required for the task.
        group_num: Number of tasks running simultaneously.
        block: If True, block allocated available GPU.
        verbosity: If True, print heartbeat while blocking.
        memory_need: Memory need for available GPU (MB). < -0.5 means need 95% GPU memory.
        max_process: Max process num on available GPU. < 0 means no limit.

    Returns:
        tuple[list of cuda indices, dict of environment variables with CUDA_VISIBLE_DEVICES set]
    """
    # -* 参数检查：group_cuda_num和group_num必须有且只有一个。
    assert (group_cuda_num is None) ^ (group_num is None), "group_cuda_num and group_num must have and only have one."

    # -* 若串行或不用卡，不做分配。
    if (group_num == 0) or (group_cuda_num == 0):
        return [], os.environ.copy()

    # -* 找到已有的CUDA设备。
    cudas = get_cudas()[0]
    cuda_num = len(cudas)  # 卡数。

    # -* 确定卡数或组数。
    if group_num is None:
        group_cuda_num: int
        group_num = cuda_num // group_cuda_num  # 卡数已知，算出组数。
    else:
        group_num: int
        group_cuda_num = cuda_num // group_num  # 组数已知，算出卡数。

    # -* 等待当前CUDA设备空闲。
    # -** 按照索引确定CUDA设备。
    group_idx = group_rank % group_num  # 当前所在组。
    current_cudas = cudas[group_idx * group_cuda_num:(group_idx + 1) * group_cuda_num]  # 所在组拥有的CUDA设备。
    # -** 动态分配。
    if block:
        current_cudas = block_get_available_cuda(cudas,
                                                 cuda_need=group_cuda_num,
                                                 memory_need=memory_need,
                                                 max_process=max_process,
                                                 verbosity=verbosity,
                                                 cudas_prefer=current_cudas)

    # -* 获取当前指定了当前设备的环境变量。
    env_with_current_cuda = os.environ.copy()
    env_with_current_cuda['CUDA_VISIBLE_DEVICES'] = cudas2CUDA_VISIBLE_DEVICES(current_cudas)

    # -* 返回分配的cuda设备。
    return current_cudas, env_with_current_cuda



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/7/22 18:17
@File    : sub_augers.py
@Software: PyCharm
@Desc    : 
"""
from typing import Union, Iterable

import numpy as np

from alchemy_cat.dag import Graph
from alchemy_cat.data.plugins.augers import RandCrop, pad_img_label, int_img2float32_img, centralize

__all__ = ['float_centralize', 'pad_crop']


def float_centralize(mean: Union[int, float, Iterable], std: Union[int, float, Iterable, None]=None,
                     scale: bool=False) -> Graph:
    sub_g = Graph(slim=True)

    if scale:
        mean = np.array(mean) / 255.
        std = np.array(std) / 255. if std is not None else None

    sub_g.add_node(int_img2float32_img, args=['img', {'scale': scale}], outputs=['float_img'])
    sub_g.add_node(centralize, args=['float_img', {'mean': mean}, {'std': std}],
                   outputs=['centralized_img'])

    return sub_g


def pad_crop(crop_size: Union[int, Iterable[int]]=321, ignore_label: int=255) -> Graph:
    sub_g = Graph(slim=True)

    sub_g.add_node(pad_img_label,
                   args=['img', 'label'],
                   kwargs={'pad_img_to': crop_size, 'pad_location': 'center', 'ignore_label': ignore_label},
                   outputs=['padded_img', 'padded_label'])

    rand_crop = RandCrop(crop_size)
    sub_g.add_node(rand_crop, args=['padded_img', 'padded_label'], outputs=['cropped_img', 'cropped_label'])

    return sub_g



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: cam_sal_to_seed.py
@time: 2020/3/27 1:10
@desc:
"""
import numpy as np
from alchemy_cat.contrib.tasks.wsss.seeding.e2e_seeding.resolve_loc_cue_conflict \
    import resolve_loc_cue_conflict_by_area_order

__all__ = ["cam_sal_to_seed"]


def cam_sal_to_seed(cam, sal, cls_in_label, cam_thresh, sal_thresh, ignore_label) -> np.ndarray:
    """Get localization cues with method in SEC paper

    Perform hard threshold for each foreground class

    Args:
        cam: (H, W, num_class - 1) cam
        sal: (H, W) saliency map
        cls_in_label: list of foreground classes
        cam_thresh: hard threshold to extract foreground class cues
        sal_thresh: hard threshold to extract background class cues
        ignore_label: ignore label in class cues

    Returns:
        (H, W) seed
    """
    loc_cue_proposal = np.zeros(shape=(cam.shape[0], cam.shape[1], cam.shape[2] + 1), dtype=np.int64)  # (H, W, num_class)
    for cls_idx in range(1, len(cls_in_label)):
        if cls_in_label[cls_idx] == 1:
            heat_map = cam[:, :, cls_idx - 1]
            loc_cue_proposal[:, :, cls_idx] = heat_map > cam_thresh * np.amax(heat_map)

    if cls_in_label[0] == 1:
        loc_cue_proposal[:, :, 0] = sal < sal_thresh

    # handle conflict seed
    seed = resolve_loc_cue_conflict_by_area_order(loc_cue_proposal, ignore_label, train_boat=True)

    return seed



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/8/10 20:07
@File    : viz_voc_annotations.py
@Software: PyCharm
@Desc    :
"""
import argparse
import os
import os.path as osp

import cv2
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from tqdm import tqdm

from alchemy_cat.py_tools import find_img_files, find_files_by_exts
from alchemy_cat.contrib.build_dt.voc_object_detection import load_voc_annotation


parser = argparse.ArgumentParser()
parser.add_argument('-r', '--root', type=str)
parser.add_argument('-l', '--label_source', default=1, type=int)
parser.add_argument('-s', '--show_viz', default=0, type=int)
args = parser.parse_args()

# * Colors
kColors = ['red', 'blue', 'darkorange', 'violet', 'cyan', 'green', 'black']

# * Create VOC Dir & Source Dir
kImgDir = osp.join(args.root, 'JPEGImages')
kAnnoDir = osp.join(args.root, 'Annotations')
kVizDir = osp.join(args.root, 'VizAnnotations')

os.makedirs(kVizDir, exist_ok=True)


# * Find all image, label, source files.
img_files = find_img_files(kImgDir)
anno_files = [osp.join(kAnnoDir, file) for file in sorted(list(find_files_by_exts(kAnnoDir, exts=['.xml'])))]

# * Prepare figure to show
fig: plt.Figure = plt.figure(dpi=150)

# * For each img, create corresponding xml annotations
for img_file, anno_file in tqdm(zip(img_files, anno_files),
                                desc='Processing', unit='imgs', dynamic_ncols=True,
                                total=len(img_files)):
    # * Get img id
    img_id = osp.splitext(osp.basename(img_file))[0]

    # * Get RGB img
    img = cv2.imread(img_file, cv2.IMREAD_COLOR)[..., ::-1]

    # * Get annotation
    anno = load_voc_annotation(anno_file)

    # * Plot bottom img
    ax: plt.Axes = fig.add_subplot()
    ax.imshow(img)
    ax.axis("off")

    # * Plot bbox
    for i, obj in enumerate(anno['objects']):
        name = obj['name']
        x_min, y_min, x_max, y_max = obj['bndbox']
        ax.add_patch(patches.Rectangle((x_min, y_min), x_max - x_min, y_max - y_min,
                                       fill=False, edgecolor=kColors[i % len(kColors)], lw=1))
        ax.text(x_min, (y_min-15), name, verticalalignment='top', color=kColors[i % len(kColors)], fontsize=5)

    # * Label source
    if args.label_source:
        ax.text(0, 0, f"{anno['source_database']} {anno['source_image']}",
                verticalalignment='top', color='darkred', fontsize=5, weight='bold')

    # * Show viz
    if args.show_viz:
        fig.show()

    # * Save viz
    fig.savefig(osp.join(kVizDir, f'{img_id}.png'), bbox_inches='tight', pad_inches=0)

    fig.clf()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/6 22:09
@File    : cfg.py.py
@Software: PyCharm
@Desc    : 
"""
from alchemy_cat.py_tools import Cfg2Tune

cfg = config = Cfg2Tune()

cfg.rslt_dir = 'no_param_config'

# * Param on same Cfg2Tune
cfg.foo0.foo1.a = 1
cfg.foo0.foo1.b = .1
# * Param on diff Cfg2Tune
cfg.foo0.fix = '1'
# * Param on root Cfg2Tune
cfg.c = True
# * Static Value
cfg.foo0.a_max = 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: figure_wall.py
@time: 2020/1/7 22:18
@desc:
"""
import os
import copy
import numpy as np
from matplotlib import pyplot as plt
from typing import Union, Optional, Iterable
from collections import abc

from alchemy_cat.acplot.utils import stack_figs
from alchemy_cat.acplot.shuffle_ch import BGR2RGB, RGB2BGR
from alchemy_cat.py_tools.type import is_intarr
from alchemy_cat.py_tools.str_formatters import indent
from alchemy_cat.alg import color2scalar


__all__ = ["SquareFigureWall", "RectFigureWall", "RowFigureWall", "ColumnFigureWall"]


KDefaultPadValue = (127, 140, 141) # BGR Default pad value


class FigureWall(object):
    """A class to operate a figure wall"""

    def __init__(self, figs: Union[np.ndarray, Iterable[Union[np.ndarray, 'FigureWall']]],
                 is_normalize: bool = False, space_width: int = 1,
                 img_pad_val: Union[int, float, Iterable] = None,
                 pad_location: Union[str, int]='right-bottom', color_channel_order: str='RGB'):
        """A class to operate a figure wall

        Args:
            figs (Iterable, np.ndarray): Iterable[fig] or figs((N, H, W, C) ndarray). fig is supposed to be (H, W, C)
                ndarray or a figure wall which will be seen as it's tiled_figs.
            is_normalize (bool): (Deprecated)If true, the figures will be min-max normalized.
            space_width (int): Space width between figs
            img_pad_val: (Union[int, float, Iterable]): Img pad value when stack imgs.If value is int or float, return
                (value, value, value), if value is Iterable with 3 element, return totuple(value), else raise error.
                (Default: BGR(127, 140, 141))
            pad_location (Union[str, int]): Img pad location when stack imgs. Indicate pad location. Can be
                'left-top'/0, 'right-top'/1, 'left-bottom'/2, 'right-bottom'/3, 'center'/4.
            color_channel_order (str): Indicate color channel order. If 'BGR', then fig of figs should be BGR ndarray. If fig is a
                FigureWall object, fig.BGR() will be called to make figure_wall's color channel order meet with
                indicated order. Same to the 'RGB'.
        """
        if color_channel_order not in ('BGR', 'RGB'):
            raise ValueError(f"color_channel_order {color_channel_order} must be \"BGR\" or \"RGB\"")
        color_mode_converter = (lambda x: x.BGR()) if color_channel_order == 'BGR' else (lambda x: x.RGB())

        if img_pad_val is None:
            img_pad_val = KDefaultPadValue if color_channel_order == 'BGR' else KDefaultPadValue[::-1]
        else:
            img_pad_val = color2scalar(img_pad_val)

        if isinstance(figs, list):
            figs_to_stack = []
            for fig in figs:
                if isinstance(fig, np.ndarray):
                    figs_to_stack.append(fig)
                elif isinstance(fig, FigureWall):
                    figs_to_stack.append(color_mode_converter(fig).tiled_figs)
                else:
                    raise ValueError(f"The fig {fig} should be ndarray or FigureWall")

            self.figs = stack_figs(figs_to_stack, img_pad_val, pad_location)
        elif isinstance(figs, np.ndarray):
            self.figs = figs
        elif isinstance(figs, abc.Iterable):
            self.__init__(list(figs), is_normalize, space_width, img_pad_val, pad_location, color_channel_order)
            return
        else:
            raise ValueError("The figs should be Iterator of (H, W, C) imgs or (N, H, W, C) ndarray")

        self.color_channel_order = color_channel_order
        self.space_width = space_width
        self.is_normalize = is_normalize
        self.img_pad_val = img_pad_val
        self.pad_location = pad_location

        if is_normalize:
            self.figs = (self.figs - self.figs.min()) / (self.figs.max() - self.figs.min())

        self.tiled_figs = self._tile_figs(self.space_width)


    def _tile_figs(self, space_width) -> np.ndarray:
        """Tiles figs to figure wall.
        Args:
            space_width (int): Space width between figs

        Returns:
            ndarray with shape (H, W, C)
        """
        raise NotImplementedError

    def plot(self, **kwargs) -> plt.Figure:
        """Plot figure wall with pyplot.
        Args:
            **kwargs (): kwargs for plt.figure()

        Returns:
            Shown figure
        """
        tiled_figs = self.tiled_figs if self.color_channel_order == 'RGB' else BGR2RGB(self.tiled_figs)

        figure = plt.figure(**kwargs)
        plt.imshow(tiled_figs)
        plt.axis('off')

        return figure

    def save(self, img_file: str, **kwargs):
        """Save figure wall to file

        Args:
            img_file (str): Path to save the figure wall
            kwargs (dict): key-word arguments for plt.imsave

        Returns:
            None
        """
        tiled_figs = self.tiled_figs if self.color_channel_order == 'RGB' else BGR2RGB(self.tiled_figs)

        os.makedirs(os.path.split(img_file)[0], exist_ok=True)

        plt.imsave(img_file, tiled_figs, **kwargs)

    def __add__(self, other: 'FigureWall') -> 'FigureWall':
        """Collect figs from self and other's figs and tile them to a new figure wall.

        Args:
            other (FigureWall): A figure wall instance

        Returns:
            A new figure wall with figs from self and other. The type and attribute of figure wall
                will follow the first factor.
        """
        if not isinstance(other, FigureWall):
            raise ValueError(f"Adder {other} <{type(other)}> should be FigureWall")

        other = other.BGR() if self.color_channel_order == 'BGR' else other.RGB()
        figs = list(self.figs) + list(other.figs)
        return type(self)(figs, self.is_normalize, self.space_width, self.img_pad_val, self.pad_location,
                          self.color_channel_order)

    def __repr__(self):
        return f"FigureWall <{self.__class__.__name__}:>" + "\n" + \
                indent(f"#figs: {self.figs.shape[0]}") + "\n" + \
                indent(f"space width: {self.space_width}")

    def BGR(self, inplace_if_converted: bool=False) -> 'FigureWall':
        """Convert to a BGR figure wall

        If figure wall is already an BGR figure wall, then return self. Else return a copy of self which is converted
        to BGR mode when inplace_if_concerted=False, or just convert figure wall self to BGR mode then return it's self
        when inplace_if_concerted=True

        See Also:
            FigureWall.RGB

        Args:
            inplace_if_converted: When figure wall is in RGB mode, then convert a copy of self to BGR mode to return if
                inplace_if_converted=False, else convert self to BGR mode then return self if replace_if_converted=True.
                (Default: False)

        Returns:
            A BGR figure wall
        """
        if self.color_channel_order == 'BGR':
            return self

        if inplace_if_converted:
            ret = self
        else:
            ret = copy.deepcopy(self)

        ret.color_channel_order = 'BGR'
        ret.figs = RGB2BGR(ret.figs)
        ret.tiled_figs = RGB2BGR(ret.tiled_figs)
        ret.img_pad_val = ret.img_pad_val[::-1]

        return ret

    def RGB(self, inplace_if_converted: bool=False) -> 'FigureWall':
        """Convert to a RGB figure wall

        If figure wall is already an RGB figure wall, then return self. Else return a copy of self which is converted
        to RGB mode when inplace_if_concerted=False, or just convert figure wall self to RGB mode then return it's self
        when inplace_if_concerted=True

        See Also:
            FigureWall.BGR

        Args:
            inplace_if_converted: When figure wall is in BGR mode, then convert a copy of self to RGB mode to return if
                inplace_if_converted=False, else convert self to RGB mode then return self if replace_if_converted=True.
                (Default: False)

        Returns:
            A RGB figure wall
        """
        if self.color_channel_order == 'RGB':
            return self

        if inplace_if_converted:
            ret = self
        else:
            ret = copy.deepcopy(self)

        ret.color_channel_order = 'RGB'
        ret.figs = BGR2RGB(ret.figs)
        ret.tiled_figs = BGR2RGB(ret.tiled_figs)
        ret.img_pad_val = ret.img_pad_val[::-1]

        return ret


class SquareFigureWall(FigureWall):
    """Figures will be tiled to an Square"""

    def _tile_figs(self, space_width):
        n = int(np.ceil(np.sqrt(self.figs.shape[0])))
        padding = (((0, n ** 2 - self.figs.shape[0]),
                    (0, space_width), (0, space_width))  # add some space between filters
                   + ((0, 0),) * (self.figs.ndim - 3))  # don't pad the last dimension (if there is one)
        if is_intarr(self.figs):
            constant_values = 255
        else:
            constant_values = 1.

        tiled_figs = np.pad(self.figs, padding, mode='constant', constant_values=constant_values)  # pad with white

        # tile the filters into an image
        tiled_figs = tiled_figs.reshape((n, n) + tiled_figs.shape[1:]).transpose(
            (0, 2, 1, 3) + tuple(range(4, tiled_figs.ndim + 1)))
        tiled_figs = tiled_figs.reshape((n * tiled_figs.shape[1], n * tiled_figs.shape[3]) + tiled_figs.shape[4:])

        return tiled_figs[:-space_width, :-space_width, ...]  # Delete the padding border


class RectFigureWall(FigureWall):
    """Figures will be tiled to a Rectangle"""

    def __init__(self, figs: Union[Iterable[Union[np.ndarray, 'FigureWall']], np.ndarray], is_normalize: bool = False,
                 space_width: int = 1, row_num: Optional[int]=None, col_num: Optional[int]=None,
                 img_pad_val: Union[int, float, Iterable] = None,
                 pad_location: Union[str, int]='right-bottom', color_channel_order: str='RGB'):
        """Figures will be tiled to an Rectangle

        Args:
            figs (list, np.ndarray): List[fig] or figs. fig is supposed to be (H, W, C) and RGB mode.
            is_normalize (bool): If true, the figures will be min-max normalized
            space_width (int): Space width between figs
            row_num (Optional[int]): row num of rectangle figure wall. If None, it will be calculated by
                ceil(figure's num / col_num)
            col_num (Optional[int]): cow num of rectangle figure wall. If None, it will be calculated by
                ceil(figure's num / row_num)
            img_pad_val: Same to the param of FigureWall
            pad_location (Union[str, int]): Same to the param of FigureWall
            color_channel_order (str): Same to the param of FigureWall

        See Also:
            FigureWall
        """
        self.row_num = row_num
        self.col_num = col_num
        super(RectFigureWall, self).__init__(figs, is_normalize, space_width,
                                             img_pad_val, pad_location, color_channel_order)

    def _tile_figs(self, space_width):
        if self.row_num is None and self.col_num is None:
            raise ValueError(f"row_num={self.row_num} or col_num={self.col_num} should not be int, not None")

        figs_num = self.figs.shape[0]
        if self.row_num is None:
            self.row_num = int(np.ceil(figs_num / self.col_num))
        elif self.col_num is None:
            self.col_num = int(np.ceil(figs_num / self.row_num))

        n = self.row_num * self.col_num

        if n < self.figs.shape[0]:
            raise ValueError(f"row_num * col_num = {n} < self.figs.shape[0]")

        padding = (((0, n - self.figs.shape[0]),
                    (0, space_width), (0, space_width))  # add some space between filters
                   + ((0, 0),) * (self.figs.ndim - 3))  # don't pad the last dimension (if there is one)
        if is_intarr(self.figs):
            constant_values = 255
        else:
            constant_values = 1.

        tiled_figs = np.pad(self.figs, padding, mode='constant', constant_values=constant_values)  # pad with white

        # tile the filters into an image
        tiled_figs = tiled_figs.reshape((self.row_num, self.col_num) + tiled_figs.shape[1:]).transpose(
            (0, 2, 1, 3) + tuple(range(4, tiled_figs.ndim + 1)))
        tiled_figs = tiled_figs.reshape((self.row_num * tiled_figs.shape[1], self.col_num * tiled_figs.shape[3]) + tiled_figs.shape[4:])

        return tiled_figs[:-space_width, :-space_width, ...]  # Delete the padding border


class ColumnFigureWall(FigureWall):
    """Figures will be tiled to a Column"""

    def _tile_figs(self, space_width):
        padding = ( ((0, 0), (0, space_width)) + ((0, 0),) * (self.figs.ndim - 2) )
        if is_intarr(self.figs):
            constant_values = 255
        else:
            constant_values = 1.

        tiled_figs = np.pad(self.figs, padding, mode='constant', constant_values=constant_values)

        return tiled_figs.reshape((tiled_figs.shape[0] * tiled_figs.shape[1],) + tiled_figs.shape[2:])[:-space_width, ...]


class RowFigureWall(FigureWall):
    """
    Figures will be tiled to a Row
    """

    def _tile_figs(self, space_width):
        padding = ( ((0, 0), (0, 0), (0, space_width)) + ((0, 0),) * (self.figs.ndim - 3) )
        if is_intarr(self.figs):
            constant_values = 255
        else:
            constant_values = 1.

        tiled_figs = np.pad(self.figs, padding, mode='constant', constant_values=constant_values)

        tiled_figs = tiled_figs.transpose((1, 0, 2) + tuple(range(3, tiled_figs.ndim)))
        tile_figs = tiled_figs.reshape((tiled_figs.shape[0], tiled_figs.shape[1] * tiled_figs.shape[2]) + tiled_figs.shape[3:])

        return tile_figs[:, :-space_width, ...]


if __name__ == "__main__":
    from alchemy_cat.contrib.voc import VOCAug
    voc_dataset = VOCAug()
    img_wall = RowFigureWall([img for _, img, label in voc_dataset[:3]], pad_location='center')
    img_wall.plot(dpi=600)

    col_wall = ColumnFigureWall(list(np.random.randn(10, 224, 224, 3)), is_normalize=True, space_width=5)
    col_wall.plot(dpi=300)

    row_wall = RowFigureWall(list(np.random.randn(10, 224, 224, 3)), is_normalize=True, space_width=5)
    row_wall.plot(dpi=300)

    sq_wall = SquareFigureWall(list(np.random.randn(10, 224, 224, 3)), is_normalize=True, space_width=5)
    sq_wall.plot(dpi=300)

    rect_wall = RectFigureWall(list(np.random.randn(10, 224, 224, 3)), is_normalize=True, space_width=5, row_num=2, col_num=5)
    rect_wall.plot(dpi=300)

    rect_wall = RectFigureWall(list(np.random.randn(9, 224, 224, 3)), is_normalize=True, space_width=5, row_num=2)
    rect_wall.plot(dpi=300)

    rect_wall = RectFigureWall(list(np.random.randn(9, 224, 224, 3)), is_normalize=True, space_width=5, col_num=4)
    rect_wall.plot(dpi=300)

    wall_wall = RowFigureWall([col_wall, row_wall, sq_wall], space_width=10)
    wall_wall.plot(dpi=300)

    add_wall = sq_wall + col_wall + row_wall
    add_wall.plot(dpi=300)

    plt.show()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: utils.py
@time: 2020/2/5 2:30
@desc:
"""
import numpy as np

__all__ = ["VOC_COLOR", "VOC_CLASSES", "color_map2label_map", "label_map2color_map", ]

VOC_COLOR = np.array([[0, 0, 0], [128, 0, 0], [0, 128, 0], [128, 128, 0],
                      [0, 0, 128], [128, 0, 128], [0, 128, 128], [128, 128, 128],
                      [64, 0, 0], [192, 0, 0], [64, 128, 0], [192, 128, 0],
                      [64, 0, 128], [192, 0, 128], [64, 128, 128], [192, 128, 128],
                      [0, 64, 0], [128, 64, 0], [0, 192, 0], [128, 192, 0],
                      [0, 64, 128]], dtype=np.uint8)

VOC_CLASSES = ['background', 'aeroplane', 'bicycle', 'bird', 'boat',
               'bottle', 'bus', 'car', 'cat', 'chair', 'cow',
               'diningtable', 'dog', 'horse', 'motorbike', 'person',
               'potted plant', 'sheep', 'sofa', 'train', 'tv/monitor']

color_id2label = np.zeros(256 ** 3, dtype=np.uint8)
for i, color in enumerate(VOC_COLOR):
    color_id2label[(color[0] * 256 + color[1]) * 256 + color[2]] = i

label2color = np.ones((256, 3), dtype=np.uint8) * 255
label2color[:21, :] = VOC_COLOR


def color_map2label_map(color_map: np.ndarray) -> np.ndarray:
    """Convert VOC color map to label map

    Args:
        color_map (np.ndarray):
            Labels with shape (..., C)
    Returns (np.ndarray):
        Labels with shape (...)
    """
    color_id_map = ((color_map[..., 0] * 256 + color_map[..., 1]) * 256
                    + color_map[..., 2])
    return color_id2label[color_id_map]


def label_map2color_map(label_map: np.ndarray) -> np.ndarray:
    """Convert VOC label map to color map

    Args:
        label_map (np.ndarray):
            Labels with shape (...)
    Returns (np.ndarray):
        color_map with shape (..., C)
    """
    color_map = label_map[..., np.newaxis]
    color_map = np.concatenate((color_map, color_map, color_map), axis=-1)

    color_map[..., 0] = label2color[..., 0][color_map[..., 0]]
    color_map[..., 1] = label2color[..., 1][color_map[..., 1]]
    color_map[..., 2] = label2color[..., 2][color_map[..., 2]]

    return color_map



# -- tune_train.py --
import argparse, json, os, subprocess, sys
from alchemy_cat.dl_config import Config, Cfg2TuneRunner

parser = argparse.ArgumentParser(description='Tuning AlchemyCat MNIST Example')
parser.add_argument('-c', '--cfg2tune', type=str)
args = parser.parse_args()

# Will run `torch.cuda.device_count() // work_gpu_num`  of configs in parallel
runner = Cfg2TuneRunner(args.cfg2tune, experiment_root='/tmp/experiment', work_gpu_num=1)

@runner.register_work_fn  # How to run config
def work(pkl_idx: int, cfg: Config, cfg_pkl: str, cfg_rslt_dir: str, cuda_env: dict[str, str]) -> ...:
    subprocess.run([sys.executable, 'train.py', '-c', cfg_pkl], env=cuda_env)

@runner.register_gather_metric_fn  # How to gather metric for summary
def gather_metric(cfg: Config, cfg_rslt_dir: str, run_rslt: ..., param_comb: dict[str, tuple[..., str]]) -> dict[str, ...]:
    return json.load(open(os.path.join(cfg_rslt_dir, 'eval.json')))

runner.tuning()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: quick_init.py
@time: 2019/9/3 1:52
@desc:
"""
from inspect import signature
from collections import OrderedDict


def quick_init(obj, local_var_dict: dict, update_to_obj_dict: bool=False) -> OrderedDict:
    """Quick init the object with all it's __init__ function's params. Use like quick_init(self, locals())

    Args:
        obj (): Object to be initialized
        local_var_dict (): locals() ->
        update_to_obj_dict (): If true, the param dict will be update to object's __dict__

    Returns:
        Dict of __init__ function's param
    """
    # init_dict_keys = obj.__init__.__code__.co_varnames[1:obj.__init__.__code__.co_argcount]
    init_dict_keys = signature(obj.__init__).parameters.keys()

    init_dict = OrderedDict()
    for key in init_dict_keys:
        init_dict[key] = local_var_dict[key]

    # Not Ordered if python version < 3.6
    # init_dict = {key: local_var_dict[key] for key in init_dict_keys}

    if update_to_obj_dict:
        obj.__dict__.update(init_dict)

    return init_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/8/2 21:09
@File    : py2_client.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np
import time

from alchemy_cat.contrib.shm_send_recv.py2_send_recv import client_send, client_recv  # 导入client所需函数

kIterTimes = 1000
kImageSize = (512, 1024)
kClientImg = np.random.randint(0, 256, size=kImageSize + (3,), dtype=np.uint8)


total_time_start = time.time()

for i in range(kIterTimes):
    sample_time_start = time.time()
    client_send(kClientImg)  # 发送图片
    predict = client_recv()  # 接收预测结果
    sample_time_end = time.time()
    print('sample time consuming = %fs' % (sample_time_end - sample_time_start))

total_time_end = time.time()

print('fps =', kIterTimes / (total_time_end - total_time_start))



from collections import OrderedDict, defaultdict

from torch import nn
import torch.nn.functional as F

from alchemy_cat.torch_tools import Network, kParamDictType

__all__ = ['convert_to_separable_conv']


class _SimpleSegmentationModel(Network):
    def __init__(self, backbone, head):
        super(_SimpleSegmentationModel, self).__init__()
        self.backbone = backbone
        self.head = head

    def forward(self, x):
        input_shape = x.shape[-2:]
        features = self.backbone(x)
        x = self.head(features)
        x = F.interpolate(x, size=input_shape, mode='bilinear', align_corners=True)
        return x

    def initialize(self, pretrain_dict: kParamDictType, verbosity: int=3, **kwargs):
        self.update_model_state_dict(pretrain_dict, verbosity)


class IntermediateLayerGetter(nn.ModuleDict, Network):
    """
    Module wrapper that returns intermediate layers from a model

    It has a strong assumption that the modules have been registered
    into the model in the same order as they are used.
    This means that one should **not** reuse the same nn.Module
    twice in the forward if you want this to work.

    Additionally, it is only able to query submodules that are directly
    assigned to the model. So if `model` is passed, `model.feature1` can
    be returned, but not `model.feature1.layer2`.

    Arguments:
        model (nn.Module): model on which we will extract the features
        return_layers (Dict[name, new_name]): a dict containing the names
            of the modules for which the activations will be returned as
            the key of the dict, and the value of the dict is the name
            of the returned activation (which the user can specify).
    """
    def __init__(self, model, return_layers):
        if not set(return_layers).issubset([name for name, _ in model.named_children()]):
            raise ValueError("return_layers are not present in model")

        orig_return_layers = return_layers
        return_layers = {k: v for k, v in return_layers.items()}
        layers = OrderedDict()
        for name, module in model.named_children():
            layers[name] = module
            if name in return_layers:
                del return_layers[name]
            if not return_layers:
                break

        super(IntermediateLayerGetter, self).__init__(layers)
        self.return_layers = orig_return_layers

    def forward(self, x):
        out = OrderedDict()
        for name, module in self.named_children():
            x = module(x)
            if name in self.return_layers:
                out_name = self.return_layers[name]
                out[out_name] = x
        return out


class AtrousSeparableConvolution(Network):
    """ Atrous Separable Convolution
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride, padding, dilation, bias):
        super(AtrousSeparableConvolution, self).__init__()
        self.body = nn.Sequential(
            # Separable Conv
            nn.Conv2d(in_channels, in_channels, kernel_size=kernel_size, stride=stride,
                      padding=padding, dilation=dilation, bias=bias, groups=in_channels),
            # PointWise Conv
            nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=1, padding=0, bias=bias),
        )

        self._init_weight()

        self.in_channel = in_channels
        self.out_channel = out_channels

    def forward(self, x):
        return self.body(x)

    def _init_weight(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)


def convert_to_separable_conv(module):
    new_module = module
    if isinstance(module, nn.Conv2d) and module.kernel_size[0] > 1:
        new_module = AtrousSeparableConvolution(module.in_channels,
                                                module.out_channels,
                                                module.kernel_size,
                                                module.stride,
                                                module.padding,
                                                module.dilation,
                                                module.bias)
    for name, child in module.named_children():
        new_module.add_module(name, convert_to_separable_conv(child))
    return new_module



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: data_manager.py
@time: 2020/3/20 23:22
@desc:
"""
from typing import Union, Optional, Callable, List, Iterable, Tuple

import os
import os.path as osp
import shutil
import pickle
import random

import numpy as np
import torch
import torch.distributed as dist
from torch.utils.data import Dataset as TorchDataset
from torch.utils.data import DataLoader, Sampler, RandomSampler, SequentialSampler, BatchSampler, DistributedSampler

from alchemy_cat.data import Prefetcher, DataAuger, Dataset, read_rand_seeds
from alchemy_cat.py_tools import indent

kBatchesType = List[List[int]]


class _IdentityMapAuger(DataAuger):
    """Auger for identity mapping"""

    def build_graph(self):
        @self.graph.register(inputs=['example'], outputs=['output'])
        def identity_mapping(example):
            return example


class _EpochBatchSampler(Sampler):
    """Batch sampler for DataLoader in DataManager"""

    def __init__(self, batches, epoch_iteration, data_source):
        super(_EpochBatchSampler, self).__init__(data_source)

        self.batches = batches[epoch_iteration:]

    def __iter__(self):
        return iter(self.batches)

    def __len__(self):
        return len(self.batches)


# 拷贝自https://github.com/pytorch/pytorch/blob/master/torch/utils/data/_utils/worker.py，可以使得到的numpy seed与random的
# seed错开。
def _generate_state(base_seed, worker_id):
    INIT_A = 0x43b0d7e5
    MULT_A = 0x931e8875
    INIT_B = 0x8b51f9dd
    MULT_B = 0x58f38ded
    MIX_MULT_L = 0xca01f9dd
    MIX_MULT_R = 0x4973f715
    XSHIFT = 4 * 8 // 2
    MASK32 = 0xFFFFFFFF

    entropy = [worker_id, base_seed & MASK32, base_seed >> 32, 0]
    pool = [0] * 4

    hash_const_A = INIT_A

    def hash(value):
        nonlocal hash_const_A
        value = (value ^ hash_const_A) & MASK32
        hash_const_A = (hash_const_A * MULT_A) & MASK32
        value = (value * hash_const_A) & MASK32
        value = (value ^ (value >> XSHIFT)) & MASK32
        return value

    def mix(x, y):
        result_x = (MIX_MULT_L * x) & MASK32
        result_y = (MIX_MULT_R * y) & MASK32
        result = (result_x - result_y) & MASK32
        result = (result ^ (result >> XSHIFT)) & MASK32
        return result

    # Add in the entropy to the pool.
    for i in range(len(pool)):
        pool[i] = hash(entropy[i])

    # Mix all bits together so late bits can affect earlier bits.
    for i_src in range(len(pool)):
        for i_dst in range(len(pool)):
            if i_src != i_dst:
                pool[i_dst] = mix(pool[i_dst], hash(pool[i_src]))

    hash_const_B = INIT_B
    state = []
    for i_dst in range(4):
        data_val = pool[i_dst]
        data_val = (data_val ^ hash_const_B) & MASK32
        hash_const_B = (hash_const_B * MULT_B) & MASK32
        data_val = (data_val * hash_const_B) & MASK32
        data_val = (data_val ^ (data_val >> XSHIFT)) & MASK32
        state.append(data_val)
    return state


def _check_sampler(sampler):
    if not hasattr(sampler, '__len__'):
        raise ValueError(f"sampler <{sampler.__class__}>: {sampler} should have attribute '__len__'")
    length = len(sampler)
    if not isinstance(length, int) or length <= 0:
        raise ValueError(f"len(sampler): {length} should be integer > 0")


def _check_batches_read(batches, batch_num, batch_size, drop_last):
    if not isinstance(batches, list) or not isinstance(batches[0], list) or not isinstance(batches[0][0], int):
        raise RuntimeError(f"batches read {batches} should be List[List[int]")

    if len(batches) < batch_num:
        raise RuntimeError(f"len(batches read) {len(batches)} != DataManager's batch num {batch_num}")

    if (len(batches) > 1 or drop_last) and len(batches[0]) != batch_size:
        raise RuntimeError(f"batch size of batches read {len(batches[0])} != "
                           f"DataManager's batch size {batch_size}")


def _check_batches_uniqueness(batches: kBatchesType, auger: DataAuger):
    indices = DataManager.batches2indices(batches)
    if len(indices) != len(set(indices)) and len(auger.rand_nodes) > 0:
        raise RuntimeError(f"batches generated by batch sampler can't have indices repeated when there is rand_nodes in"
                           f"data_auger's graph")


class DataManager(object):
    """A dataloader which can record and recover the process of loading"""

    __initialized: bool = False

    def __init__(self, dataset: Union[Dataset, TorchDataset, None]=None, data_auger: Optional[DataAuger]=None,
                 log_dir: str='.', is_prefetch: bool=False, log_rand_seeds: bool=True,
                 batch_size: int=1, shuffle: bool=False, sampler: Optional[Sampler]=None,
                 batch_sampler: Optional[Sampler]=None, num_workers: int=0, collate_fn: Optional[Callable]=None,
                 pin_memory: bool=False, drop_last: bool=False, timeout: int=0,
                 worker_init_fn: Optional[Callable]=None,
                 generator: Optional[torch.Generator]=None, prefetch_factor: int=2):
        """ A dataloader which can record and recover the process of loading

        Args:
            dataset: Dataset to be loaded. Can be gotten from data_auger
            data_auger: DataAuger to be loaded. If None, will create an identity mapping auger with dataset.
            log_dir: Dictionary where DataManager save its log
            is_prefetch: If True, data loader iter will be wrapped by Prefetcher, which can overlap the data transfer
                and calculating on GPU. Only usable when cuda is available. (Default: False)
            log_rand_seeds: If True, rand seeds will be recorded. (Default: True)
            batch_size: Same to param for torch.data.DataLoader
            shuffle: Same to param for torch.data.DataLoader
            sampler: Same to param for torch.data.DataLoader
            batch_sampler: Same to param for torch.data.DataLoader
            num_workers: Same to param for torch.data.DataLoader
            collate_fn: Same to param for torch.data.DataLoader
            pin_memory: Same to param for torch.data.DataLoader
            drop_last: Same to param for torch.data.DataLoader
            timeout: Same to param for torch.data.DataLoader
            worker_init_fn: Same to param for torch.data.DataLoader
            generator: Same to param for torch.data.DataLoader
            prefetch_factor: Same to param for torch.data.DataLoader

        See Also:
            torch.utils.data.DataLoader: https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader
        """
        if dataset is None and data_auger is None:
            raise ValueError(f"dataset and data_auger can't be None at the same time")

        if dataset is not None and data_auger is not None:
            if data_auger.dataset is not dataset:
                raise ValueError(f"data_auger.dataset {data_auger.dataset} should be dataset {dataset}")

        if dataset is None:
            dataset = data_auger.dataset
        elif data_auger is None:
            data_auger = _IdentityMapAuger(dataset, slim=True)

        self._dataset = dataset
        self._data_auger = data_auger

        self.log_dir = log_dir
        self.is_prefetch = is_prefetch
        self.log_rand_seeds = log_rand_seeds

        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.timeout = timeout
        self.worker_init_fn = worker_init_fn
        self.collect_fn = collate_fn
        self.generator = generator
        self.prefetch_factor = prefetch_factor

        self.shuffle = shuffle
        self.sampler = sampler
        self.drop_last = drop_last
        self.batch_size = batch_size
        self.batch_sampler = batch_sampler

        if self.batch_sampler is not None:
            if self.shuffle or self.sampler is not None or self.drop_last or self.batch_size != 1:
                raise ValueError('batch_sampler option is mutually exclusive '
                                 'with batch_size, shuffle, sampler, and '
                                 'drop_last')
            _check_sampler(self.batch_sampler)
            self.batch_size = len(next(iter(self.batch_sampler)))
            self.shuffle = None  # UnKnown
            self.drop_last = None  # UnKnown
        else:
            if self.sampler is not None:
                if self.shuffle:
                    raise ValueError('sampler option is mutually exclusive with shuffle')
                _check_sampler(self.sampler)
                self.shuffle = None  # UnKnown
            elif self.shuffle:
                self.sampler = RandomSampler(self.data_source, generator=self.generator)
            else:
                self.sampler = SequentialSampler(self.data_source)

            self.batch_sampler = BatchSampler(self.sampler, batch_size, drop_last)

        self._epoch: int = -1
        self.epoch_batches: Optional[kBatchesType] = None
        self.epoch_loader: Optional[DataLoader] = None
        self.epoch_iter: Optional[Iterable] = None

        def worker_init_fn_(worker_id):
            seed = torch.initial_seed()
            random.seed(seed)
            np_seed = _generate_state(seed - worker_id, worker_id)
            np.random.seed(np_seed)
            if worker_init_fn is not None:
                worker_init_fn(worker_id)

        self.worker_init_fn = worker_init_fn_

        self.__initialized: bool = True

    @property
    def dataset(self):
        return self._dataset

    @property
    def data_auger(self):
        return self._data_auger

    @property
    def auger(self):
        return self._data_auger

    @property
    def data_source(self):
        return self._data_auger

    @property
    def log_parent_dir(self):
        if dist.is_initialized():
            log_parent_dir = osp.join(self.log_dir, '.data_manager_log', f'rank{dist.get_rank()}')
        else:
            log_parent_dir = osp.join(self.log_dir, '.data_manager_log')
        return log_parent_dir

    def epoch_log_dir(self, epoch):
        return osp.join(self.log_parent_dir, f"epoch-{epoch}")

    def batches_log_file(self, epoch):
        return osp.join(self.epoch_log_dir(epoch), 'batches.pkl')

    def rand_seed_log_file(self, epoch):
        return osp.join(self.epoch_log_dir(epoch), 'rand_seed_log')

    def __setattr__(self, attr, val):
        if self.__initialized and attr in ('batch_size', 'batch_sampler', 'sampler', 'drop_last', '_dataset'
                                           '_data_auger', 'shuffle'):

            raise ValueError('{} attribute should not be set after {} is '
                             'initialized'.format(attr, self.__class__.__name__))

        super(DataManager, self).__setattr__(attr, val)

    @staticmethod
    def batches2indices(batches: kBatchesType):
        """Flat batches to indices"""
        return [index for batch in batches for index in batch]

    @property
    def batch_num(self):
        """Number of batches in each epoch"""
        return len(self.batch_sampler)

    def __len__(self):
        """Number of batches in each epoch"""
        return self.batch_num

    @property
    def epoch(self):
        """DataManager's current epoch"""
        return self._epoch

    @property
    def epoch_iteration(self):
        """DataManager's current epoch iteration. Need PyTorch >= 1.4"""
        return self.epoch_iter._num_yielded

    @property
    def epoch_indices(self):
        return self.batches2indices(self.epoch_batches) if self.epoch_batches is not None else None

    def read_batches(self, epoch: int) -> Optional[kBatchesType]:
        """Return batches of specified epoch

        Args:
            epoch: epoch num

        Returns:
            If epoch record exits, return batches. Else return None.
        """
        if not osp.isfile(self.batches_log_file(epoch)):
            return None

        with open(self.batches_log_file(epoch), 'rb') as batches_log:
            batches: kBatchesType = pickle.load(batches_log)

        _check_batches_read(batches, self.batch_num, self.batch_size, self.drop_last)
        return batches

    def save_batches(self, epoch, batches):
        with open(self.batches_log_file(epoch), 'wb') as batches_log:
            pickle.dump(batches, batches_log)

    def read_rand_seeds(self, epoch, idx):
        return read_rand_seeds(self.rand_seed_log_file(epoch), idx, create=False)

    def locate_epoch(self, iteration: int) -> Tuple[int, int]:
        """Return epoch and epoch_iteration according to total iteration"""
        return iteration // self.batch_num, iteration % self.batch_num

    def backtrace(self, iteration: Optional[int] = None,
                  epoch: Optional[int] = None, epoch_iteration: Optional[int] = None,
                  epoch_loc: Optional[int] = None, iteration_loc: Optional[int] = None) -> dict:
        """Return the backtrace of specified sample

        Args:
            iteration: The total iteration where the sample at
            epoch: The epoch where the sample at
            epoch_iteration: The epoch iteration where the sample at
            epoch_loc: Number of sample in its epoch's indices
            iteration_loc: Number of sample in its iteration's indices

        Returns:
            A dict 'ret' with keys and values:
                ret['dataset_idx']: sample's dataset index
                ret['data_auger_idx']: sample's data_auger index
                ret['node_indices']: sample's data_auger.graph's node indices
                ret['rand_seeds']: sample's data_auger.graph's rand seeds
                ret['auger_input']: origin sample from dataset
                ret['auger_output']: data_auger's output of sample
                ret['graph_data']: sample's data_auger's calculate data
        """
        if iteration is not None:
            epoch_, epoch_iteration_ = self.locate_epoch(iteration)

            if epoch is not None and epoch != epoch_:
                raise ValueError(f"epoch calculated from iteration {epoch_} != input epoch {epoch}")
            else:
                epoch = epoch_

            if epoch_iteration is not None and epoch_iteration != epoch_iteration_:
                raise ValueError(f"epoch iteration calculated from iteration {epoch_iteration_} != "
                                 f"input epoch_iteration {epoch_iteration}")
            else:
                epoch_iteration = epoch_iteration_

        if epoch is None:
            raise ValueError(f"Either iteration or epoch needed for locating sample")

        if epoch_loc is not None:
            epoch_iteration_, iteration_loc_ = epoch_loc // self.batch_size, epoch_loc % self.batch_size

            if epoch_iteration is not None and epoch_iteration != epoch_iteration_:
                raise ValueError(f"epoch iteration calculated from epoch_loc {epoch_iteration_} != "
                                 f"input epoch_iteration {epoch_iteration}")
            else:
                epoch_iteration = epoch_iteration_

            if iteration_loc is not None and iteration_loc != iteration_loc_:
                raise ValueError(f"iteration location calculated from epoch_loc {iteration_loc_} != "
                                 f"input iteration_loc {iteration_loc}")
            else:
                iteration_loc = iteration_loc_

        if epoch_iteration is None or iteration_loc is None:
            raise ValueError(f"epoch_iteration or iteration needed to get epoch iteration."
                             f"And epoch_loc or iteration_loc needed to get iteration location")

        batches = self.read_batches(epoch)
        if batches is None:
            raise RuntimeError(f"Can't find batches log")

        auger_idx = batches[epoch_iteration][iteration_loc]

        rand_seeds = self.read_rand_seeds(epoch, auger_idx)
        if rand_seeds is None:
            raise RuntimeError(f"Can't find rand seeds log")

        self.auger.load_rand_seeds(rand_seeds)

        ret = {}
        ret['data_auger_idx'] = auger_idx
        ret['dataset_idx'], ret['node_indices'] = self.auger.calculate_indices(auger_idx)
        ret['rand_seeds'] = rand_seeds

        ret['auger_input'] = self.dataset[ret['dataset_idx']]
        ret['auger_output'] = self.auger[auger_idx]
        ret['graph_data'] = self.auger.graph.data

        return ret

    def move_epoch_to(self, epoch: int, epoch_iteration: int, recreate: bool=False):
        """Move current epoch to (epoch, epoch_iteration).

        Args:
            epoch: epoch moved to
            epoch_iteration: epoch iteration moved to
            recreate: If True, the origin record of epoch moved to will be recreated.
        """
        if isinstance(self.sampler, DistributedSampler):
            self.sampler.set_epoch(epoch)
        self._epoch = epoch

        if recreate and osp.isdir(self.epoch_log_dir(epoch)):
            shutil.rmtree(self.epoch_log_dir(epoch))

        if not osp.isdir(self.epoch_log_dir(epoch)):
            os.makedirs(self.epoch_log_dir(epoch))

        batches = self.read_batches(epoch)
        if batches is None:
            batches = list(iter(self.batch_sampler))
            _check_batches_uniqueness(batches, self.auger)
            self.save_batches(epoch, batches)
        self.epoch_batches = batches

        self.auger.rand_seed_log = self.rand_seed_log_file(epoch) if self.log_rand_seeds else None

        epoch_batch_sampler = _EpochBatchSampler(batches, epoch_iteration, self.data_source)

        if self.collect_fn is None:
            # The default value of collate_fn changing with torch version
            self.epoch_loader = DataLoader(self.data_source, batch_sampler=epoch_batch_sampler,
                                           pin_memory=self.pin_memory,
                                           num_workers=self.num_workers, timeout=self.timeout,
                                           worker_init_fn=self.worker_init_fn,
                                           generator=self.generator, prefetch_factor=self.prefetch_factor)
        else:
            self.epoch_loader = DataLoader(self.data_source, batch_sampler=epoch_batch_sampler,
                                           pin_memory=self.pin_memory, collate_fn=self.collect_fn,
                                           num_workers=self.num_workers, timeout=self.timeout,
                                           worker_init_fn=self.worker_init_fn,
                                           generator=self.generator, prefetch_factor=self.prefetch_factor)

        self.epoch_iter = iter(self.epoch_loader) if not self.is_prefetch else Prefetcher(iter(self.epoch_loader))

    def start_epoch(self, iteration: Optional[int]=None, recreate: bool=False) -> Iterable:
        """Start an epoch.

        Args:
            iteration: If not None, start epoch at specified iteration. Else start epoch behind current epoch.
                (Default: None)
            recreate: If True, the origin record of epoch starting at moved to will be recreated.

        Returns:
            Current epoch's iter(DataLoader)
        """
        if iteration is not None:
            epoch, epoch_iteration = self.locate_epoch(iteration)
            self.move_epoch_to(epoch, epoch_iteration, recreate)
        else:
            self.move_epoch_to(self.epoch + 1, 0, recreate)

        return self.epoch_iter

    def next_batch(self, recreate: bool=False):
        """Get next batch from DataManager. This function can work infinitely without raising StopIteration.

        Args:
            recreate: If True, the origin record of new epoch moved to will be recreated.

        Returns:
            next batch
        """
        if self.epoch == -1:
            self.start_epoch(recreate=recreate)

        try:
            batch = next(self.epoch_iter)
        except StopIteration:
            self.start_epoch(recreate=recreate)
            batch = next(self.epoch_iter)

        return batch

    def __iter__(self):
        return self.epoch_iter

    def __repr__(self):
        return f"DataManager <{self.__class__}>: \n" \
               + indent(f'epoch_batch_num: {len(self)}') + '\n' \
               + indent(f'batch_size: {self.batch_size}') + '\n' \
               + indent(f'shuffle: {self.shuffle}') + '\n' \
               + indent(f'drop_last: {self.drop_last}') + '\n' \
               + indent(f'num_workers: {self.num_workers}') + '\n' \
               + indent(f'pin_memory: {self.pin_memory}') + '\n' \
               + indent(f'is_prefetch: {self.is_prefetch}') + '\n' \
               + indent(f'dataset: {self.dataset}') + '\n' \
               + indent(f'data_auger: {self.data_auger}')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/3/5 14:13
@File    : voc_dt.py
@Software: PyCharm
@Desc    : 
"""
import os.path as osp
import pickle

import cv2
import numpy as np
from PIL import Image
from addict import Dict
from alchemy_cat.acplot import BGR2RGB, RGB2BGR
from alchemy_cat.data import Dataset, Subset
from alchemy_cat.dl_config import ADict
from math import ceil

__all__ = ['COCO_NAMES', 'COCO_COLOR', 'COCO91to81', 'COCO81to91', 'COCO']


COCO_NAMES = ['background',
              'person', 'bicycle', 'car', 'motorcycle', 'airplane',
              'bus', 'train', 'truck', 'boat', 'traffic light',
              'fire hydrant', 'stop sign', 'parking meter', 'bench', 'bird',
              'cat', 'dog', 'horse', 'sheep', 'cow',
              'elephant', 'bear', 'zebra', 'giraffe', 'backpack',
              'umbrella', 'handbag', 'tie', 'suitcase', 'frisbee',
              'skis', 'snowboard', 'sports ball', 'kite', 'baseball bat',
              'baseball glove', 'skateboard', 'surfboard', 'tennis racket', 'bottle',
              'wine glass', 'cup', 'fork', 'knife', 'spoon',
              'bowl', 'banana', 'apple', 'sandwich', 'orange',
              'broccoli', 'carrot', 'hot dog', 'pizza', 'donut',
              'cake', 'chair', 'couch', 'potted plant', 'bed',
              'dining table', 'toilet', 'tv', 'laptop', 'mouse',
              'remote', 'keyboard', 'cell phone', 'microwave', 'oven',
              'toaster', 'sink', 'refrigerator', 'book', 'clock',
              'vase', 'scissors', 'teddy bear', 'hair drier', 'toothbrush',
              ]

COCO_COLOR = np.array([(0, 0, 0),
                       (192, 0, 192), (64, 96, 0), (0, 0, 32), (64, 96, 64), (64, 96, 128),
                       (0, 0, 160), (0, 0, 224), (64, 0, 32), (64, 0, 96), (64, 0, 160),
                       (64, 0, 224), (192, 160, 0), (192, 160, 128), (192, 160, 192), (0, 128, 0),
                       (192, 192, 32), (0, 128, 64), (0, 128, 128), (192, 192, 160), (192, 96, 0),
                       (0, 128, 192), (64, 128, 0), (192, 128, 32), (64, 128, 64), (192, 128, 96),
                       (192, 96, 128), (64, 128, 128), (192, 128, 160), (192, 0, 32), (64, 128, 192),
                       (192, 128, 224), (0, 64, 0), (192, 0, 96), (0, 64, 64), (192, 0, 160),
                       (0, 64, 128), (192, 0, 224), (0, 64, 192), (0, 0, 0), (64, 64, 0),
                       (64, 64, 64), (0, 224, 0), (64, 64, 192), (0, 224, 128), (0, 224, 192),
                       (0, 192, 32), (0, 192, 96), (0, 192, 160), (0, 96, 0), (192, 192, 0),
                       (0, 192, 224), (64, 224, 0), (64, 192, 32), (64, 224, 64), (0, 96, 128),
                       (64, 192, 96), (64, 224, 128), (64, 192, 160), (0, 96, 192), (192, 192, 192),
                       (64, 160, 0), (0, 64, 32), (64, 160, 64), (0, 64, 96), (64, 160, 128),
                       (0, 64, 160), (64, 160, 192), (0, 64, 224), (64, 64, 32), (64, 64, 160),
                       (128, 0, 0), (192, 224, 0), (192, 224, 64), (128, 0, 128), (192, 224, 128),
                       (128, 0, 192), (0, 192, 0), (0, 192, 64), (0, 192, 128), (0, 192, 192)], dtype=np.uint8)

COCO91to81 = np.array([0,
                       1, 2, 3, 4, 5,
                       6, 7, 8, 9, 10,
                       11, 0, 12, 13, 14,
                       15, 16, 17, 18, 19,
                       20, 21, 22, 23, 24,
                       0, 25, 26, 0, 0,
                       27, 28, 29, 30, 31,
                       32, 33, 34, 35, 36,
                       37, 38, 39, 40, 0,
                       41, 42, 43, 44, 45,
                       46, 47, 48, 49, 50,
                       51, 52, 53, 54, 55,
                       56, 57, 58, 59, 60,
                       0, 61, 0, 0, 62,
                       0, 63, 64, 65, 66,
                       67, 68, 69, 70, 71,
                       72, 73, 0, 74, 75,
                       76, 77, 78, 79, 80], dtype=np.uint8)

COCO81to91 = np.array([0,
                       1, 2, 3, 4, 5,
                       6, 7, 8, 9, 10,
                       11, 13, 14, 15, 16,
                       17, 18, 19, 20, 21,
                       22, 23, 24, 25, 27,
                       28, 31, 32, 33, 34,
                       35, 36, 37, 38, 39,
                       40, 41, 42, 43, 44,
                       46, 47, 48, 49, 50,
                       51, 52, 53, 54, 55,
                       56, 57, 58, 59, 60,
                       61, 62, 63, 64, 65,
                       67, 70, 72, 73, 74,
                       75, 76, 77, 78, 79,
                       80, 81, 82, 84, 85,
                       86, 87, 88, 89, 90], dtype=np.uint8)

label2color = np.ones((256, 3), dtype=np.uint8) * 255
label2color[:81, :] = COCO_COLOR


class COCO(Dataset):
    """
    coco Segmentation base dataset
    """
    class_names = COCO_NAMES
    class_num = len(class_names)
    ignore_label = 255  # 事实上，COCO真值中不存在ignore。
    coco91to81 = COCO91to81
    coco81to91 = COCO81to91

    def __init__(self, root: str="./contrib/datasets", year="2014", split: str="train", subsplit: str="",
                 label_type: str='',
                 cls_labels_type: str | None='seg_cls_labels',
                 ps_mask_dir: str=None,
                 rgb_img: bool=False,
                 PIL_read: bool=True):
        # * 记录路径参数。
        self.root = root
        self.year = year
        assert split in ['train', 'val']
        self.split = split
        if subsplit:
            if split == 'train':
                assert subsplit in ['1250']
            if split == 'val':
                assert subsplit in ['1000', '5000']
        self.subsplit = subsplit

        # * 像素级标签选择。
        if label_type:
            assert label_type in ['wt', 'ur']
        self.label_type = label_type

        # * 图像级标签选择。
        assert cls_labels_type in ('seg_cls_labels', None)
        self.cls_labels_type = cls_labels_type

        # * 记录伪真值目录。
        self.ps_mask_dir = ps_mask_dir

        # * 是否返回RGB图像。
        self.rgb_img = rgb_img

        # * 是否用PIL读取。
        self.PIL_read = PIL_read

        # * 初始化属性。
        self.image_ids = []

        self.image_dir: str | None = None
        self.label_dir: str | None = None

        # * 读取元信息。
        self._set_files()

        # * 读取图像级标签。
        if self.cls_labels_type is not None:
            with open(osp.join(self.root, 'third_party',
                               f'{cls_labels_type}' + (f'_{self.label_type}' if self.label_type else '') + '.pkl'),
                      'rb') as pkl_f:
                self.id2cls_labels = pickle.load(pkl_f)
        else:
            self.id2cls_labels = None

    @classmethod
    def subset(cls,
               split_idx_num: tuple[int, int],
               root: str="./contrib/datasets", year="2014", split: str="train", subsplit: str="",
               label_type: str = '',
               cls_labels_type: str | None = 'seg_cls_labels',
               ps_mask_dir: str = None,
               rgb_img: bool = False,
               PIL_read: bool = True):
        dt = cls(root, year, split, subsplit, label_type, cls_labels_type, ps_mask_dir, rgb_img, PIL_read)

        split_idx, split_num = split_idx_num
        assert split_idx < split_num

        step = ceil(len(dt) / split_num)
        indexes = list(range(split_idx * step, min((split_idx + 1) * step, len(dt))))

        sub_dt = Subset(dt, indexes)
        return sub_dt

    def _set_files(self):
        self.root = osp.join(self.root, f'coco{self.year}')
        self.image_dir = osp.join(self.root, 'images', f'{self.split}{self.year}')
        self.label_dir = osp.join(self.root, 'annotations',
                                  f'{self.split}{self.year}' + (f'_{self.label_type}' if self.label_type else ''))

        set_id_file = osp.join(self.root, 'imagesLists',
                               f'{self.split}' + (f'_{self.subsplit}' if self.subsplit else '') + '.txt')
        with open(set_id_file, 'r') as f:
            self.image_ids = [id_.rstrip() for id_ in f.readlines()]

    def __len__(self):
        return len(self.image_ids)

    def get_by_img_id(self, img_id: str):
        try:
            index = self.image_ids.index(img_id)
        except ValueError:
            raise RuntimeError(f"Can't find img_id {img_id} in dataset's image_ids list")
        return self[index]

    def get_item(self, index: int) -> Dict:
        # * Load image
        img_id = self.image_ids[index]
        img_path = osp.join(self.image_dir, f'{img_id}.jpg')
        if self.PIL_read:
            img = np.asarray(Image.open(img_path), dtype=np.uint8)
            if not self.rgb_img:
                img = RGB2BGR(img).copy()
        else:
            img = cv2.imread(img_path, cv2.IMREAD_COLOR)
            if self.rgb_img:
                img = BGR2RGB(img).copy()

        # * Load Label
        lb_path = osp.join(self.label_dir, f'{img_id}.png')
        lb = np.asarray(Image.open(lb_path), dtype=np.uint8)

        # * 构造输出。
        out = ADict()
        out.img_id, out.img = img_id, img
        if self.cls_labels_type is not None:
            out.cls_lb = self.id2cls_labels[img_id]

        if self.ps_mask_dir is not None:
            out.lb = np.asarray(Image.open(osp.join(self.ps_mask_dir, f'{img_id}.png')), dtype=np.uint8)
        else:
            out.lb = lb

        return out

    @staticmethod
    def label_map2color_map(label_map: np.ndarray) -> np.ndarray:
        return label2color[label_map]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/7/23 5:01
@File    : set_network.py
@Software: PyCharm
@Desc    : 设置torch网络。
"""
from torch import nn

__all__ = ['set_bn_momentum']


def set_bn_momentum(model: nn.Module, momentum=0.1):
    """设置网络BN层的动量。

    Args:
        model: torch模型。
        momentum: 动量。
    """
    for m in model.modules():
        if isinstance(m, nn.BatchNorm2d):
            m.momentum = momentum



import pytest

from alchemy_cat.dag.errors import PyungoError
from alchemy_cat.dag.utils import get_function_return_names


def test_get_function_return_names_simple():

    def a():
        b = 2
        return b

    assert get_function_return_names(a) == ['b']


def test_get_function_return_names_no_return():

    def a():
        pass

    with pytest.raises(PyungoError) as err:
        get_function_return_names(a)

    assert str(err.value) == 'No return statement found in a'


def test_get_function_return_names_tuple():

    def a():
        b = 2
        c = 3
        return b, c

    assert get_function_return_names(a) == ['b', 'c']


def test_get_function_return_names_not_valid():

    def a():
        b = 2
        c = 3
        return b / c

    with pytest.raises(PyungoError) as err:
        get_function_return_names(a)

    msg = 'Variable name or Tuple of variable names are expected, got BinOp'
    assert str(err.value) == msg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/8/2 21:09
@File    : py2_client.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np
import time

from alchemy_cat.contrib.shm_send_recv.py2_send_recv import client_send, client_recv  # 导入client所需函数

kIterTimes = 1000
kImageSize = (512, 1024)

kClientImg = np.ones(kImageSize + (3,), dtype=np.uint8)
kClientArr = np.ones(kImageSize, dtype=np.float32)
kClientObj = list(np.arange(100))

kServerObj = {'a': 1, (3,): 2.2, 'c': 'str', 'd': [1, 2]}

time_start = time.time()

for i in range(kIterTimes):
    client_send(kClientImg)  # 发送图片
    client_send(kClientArr)  # 发送numpy array
    client_send(kClientObj)  # 发送python对象

    obj_recv = client_recv()  # 接收python对象
    assert obj_recv == kServerObj

    arr_recv = client_recv()  # 接收numpy array
    assert arr_recv.shape == kImageSize
    assert np.all(arr_recv == 255.)
    assert arr_recv.dtype == np.float32

    predict = client_recv()  # 接收预测结果
    assert predict.shape == kImageSize
    assert np.all(predict == 255)
    assert predict.dtype == np.uint8

    print('Finish %d/%d' % ((i + 1), kIterTimes))

time_end = time.time()

print('fps =', kIterTimes / (time_end - time_start))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/6/26 14:27
@File    : crf_eval_probs.py
@Software: PyCharm
@Desc    : 
"""
import argparse
import glob
import os
import os.path as osp

import numpy as np
from alchemy_cat.alg.dense_crf import ParDenseCRF
from alchemy_cat.contrib.evaluation.semantics_segmentation import eval_preds
from alchemy_cat.contrib.tasks.wsss.seeding2 import idx2seed
from alchemy_cat.contrib.voc import VOCAug2
from alchemy_cat.data.plugins import identical, arr2PIL
from alchemy_cat.dl_config import Config
from tqdm import tqdm


def loader(cam_dir: str, img_id: str, score_type: str='prob') -> (np.ndarray, np.ndarray):
    loaded = np.load(osp.join(cam_dir, f'{img_id}.npz'))
    prob = loaded[score_type]
    fg_cls = loaded['fg_cls']

    return prob, fg_cls


cfg = config = Config()

cfg.crf.ini.iter_max = 10
cfg.crf.ini.pos_w = 3
cfg.crf.ini.pos_xy_std = 1
cfg.crf.ini.bi_w = 4
cfg.crf.ini.bi_xy_std = 67
cfg.crf.ini.bi_rgb_std = 3
cfg.crf.ini.img_preprocess = identical  # RGB, uint8, (H, W, 3)图片。
cfg.crf.ini.align_corner = False
cfg.crf.ini.pool_size = 0

crf = ParDenseCRF(**cfg.crf.ini)


parser = argparse.ArgumentParser()
parser.add_argument('-s', '--source', type=str)
parser.add_argument('-t', '--target', type=str)
parser.add_argument('--crf', type=int, default=0)
args = parser.parse_args()

os.makedirs(args.target, exist_ok=True)

dt = VOCAug2('datasets', split='val', rgb_img=True)

for prob_file in tqdm(glob.glob(osp.join(args.source, '*.npz')), desc="处理", unit="张", dynamic_ncols=True):
    img_id = osp.splitext(osp.basename(prob_file))[0]
    inp = dt.get_by_img_id(img_id)
    img, lb = inp.img, inp.lb

    prob, fg_cls = loader(args.source, img_id)

    if args.crf:
        prob = crf(img, prob)

    pred = idx2seed(np.argmax(prob, axis=0), fg_cls).astype(np.uint8)

    arr2PIL(pred).save(osp.join(args.target, f'{img_id}.png'))

metric = eval_preds(class_num=dt.class_num,
                    class_names=dt.class_names,
                    preds_dir=str(args.target),
                    preds_ignore_label=dt.ignore_label,
                    gts_dir=dt.label_dir,
                    gts_ignore_label=dt.ignore_label,
                    result_dir=None,
                    importance=0,
                    eval_individually=False,
                    take_pred_ignore_as_a_cls=False)

metric.print_statistics(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/23 19:52
@File    : loc_cue.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np

__all__ = ["get_loc_cue_max_thresh", "get_loc_cue_contrib_thresh"]


def get_loc_cue_max_thresh(bg_fg_score: np.ndarray, thresh) -> np.ndarray:
    """Get localization cues according bg_fg_score.

    Args:
        bg_fg_score: (class_num, ...) score.
        thresh: hard threshold to extract cues.

    Returns:
        (class_num, ...) loc cue.
    """
    return (bg_fg_score > thresh).astype(np.uint8)


def get_loc_cue_contrib_thresh(bg_fg_score: np.ndarray, thresh) -> np.ndarray:
    """Get localization cues according bg_fg_score.

    Args:
        bg_fg_score: (class_num, ...) score.
        thresh: contrib threshold to extract cues.

    Returns:
        (class_num, ...) loc cue.
    """
    flat_score = bg_fg_score.reshape(bg_fg_score.shape[0], -1)
    loc_cue = np.zeros(shape=flat_score.shape, dtype=np.uint8)

    # * L1 Norm
    flat_score /= np.sum(flat_score, axis=1, keepdims=True)

    # * Sort
    bg_fg_argsort = np.argsort(flat_score, axis=1)[:, ::-1]
    bg_fg_sort = np.sort(flat_score, axis=1)[:, ::-1]
    bg_fg_cumsum = np.cumsum(bg_fg_sort, axis=1)

    # Set cue
    for bg_fg_idx in range(flat_score.shape[0]):
        cue_num = 0
        for score_cum_sum in bg_fg_cumsum[bg_fg_idx]:
            cue_num += 1
            if score_cum_sum > thresh:
                break

        loc_cue[bg_fg_idx, bg_fg_argsort[bg_fg_idx, :cue_num]] = 1

    return loc_cue.reshape(bg_fg_score.shape)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: logger.py
@time: 2020/3/23 21:49
@desc:
"""
import sys
import os
import os.path as osp

__all__ = ['Logger']


class _StreamLogger(object):

    def __init__(self, stream_name: str, log, real_time: bool=False, silence: bool=False):
        if stream_name == 'stdout':
            self.stream = sys.stdout
            sys.stdout = self
        elif stream_name == 'stderr':
            self.stream = sys.stderr
            sys.stderr = self
        else:
            raise ValueError(f"stream_name = {stream_name} should be 'stdout' or 'stderr'.")

        self.log = log

        self.real_time = real_time
        self.silence = silence

    def write(self, message):
        if not self.silence:
            self.stream.write(message)

        self.log.write(message)

        if self.real_time:
            self.log.flush()

    def flush(self):
        self.stream.flush()
        self.log.flush()


class Logger(object):
    """After call Logger(outfile), the stdout will print to both stdout and out_file"""

    def __init__(self, out_file, real_time: bool=False, silence: bool=False):
        """After call Logger(outfile), the stdout will print to both stdout and out_file

        Args:
            out_file: File where record the stdout
            real_time: If True, log file will flush after every write() call. (Default: False)
            silence: If True, output to terminal will be suppressed. (Default: False)
        """
        sys.alchemy_cat_logger = self  # Make Logger won't be deleted.

        os.makedirs(osp.dirname(out_file), exist_ok=True)
        self.log = open(out_file, "w", encoding="utf-8")

        self.real_time = real_time
        self.silence = silence

        self.stdout_logger = _StreamLogger('stdout', self.log, self.real_time, self.silence)
        self.stderr_logger = _StreamLogger('stderr', self.log, self.real_time, self.silence)

    def __del__(self):
        self.stdout_logger.flush()
        self.stderr_logger.flush()
        self.log.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: conf_matrix_bases_metrics.py
@time: 2020/2/26 23:16
@desc:
"""
import os
import traceback
from typing import Optional, Iterable

import numpy as np
import pandas as pd
from alchemy_cat.acplot import pretty_plot_confusion_matrix
from alchemy_cat.py_tools import Statistic, Tracker, rprint, quick_init
from matplotlib import pyplot as plt

__all__ = ['ClassificationMetric', 'SegmentationMetric']


class ClassificationMetric(Tracker):
    """Multi classification metric

    See statistics at https://en.wikipedia.org/wiki/Confusion_matrix
    """

    def __init__(self, class_num: Optional[int] = None, class_names: Optional[Iterable[str]]=None):
        # * Init attributes and save init_dict
        super(ClassificationMetric, self).__init__(quick_init(self, locals()))

        # * Check Input
        if class_names is None:
            if class_num is None:
                raise ValueError("class_num and class_names can not be None at the same time")
            else:
                class_names = [str(i) for i in range(class_num)]
        else:
            class_names = [str(name) for name in class_names]
            if class_num is not None and len(class_names) != class_num:
                raise ValueError(f"class_num = {class_num} should be equal to len(class_names) = {len(class_names)}")
            else:
                class_num = len(class_names)

        self.class_num, self.class_names = class_num, class_names

        self.conf_matrix = np.zeros((class_num, class_num), dtype=np.int64)
        self._last_conf_matrix = self.conf_matrix.copy()

    @staticmethod
    def get_conf_matrix(pred: np.ndarray, gt: np.ndarray, class_num: int) -> np.ndarray:
        """Get confusion matrix accroding to input pred and gt

        Function get pred and gt array from input, which mean predicting label and ground truth label respectively.
        The label can be class of the input image, or class of each pixel of the input image. As long as pred and gt
        have the same shape.

        Args:
            pred: Numpy array of pred label
            gt: Numpy array of ground truth label.
            class_num: number of classes

        Returns:
            confusion matrix calculated according to input pred and gt
        """
        if pred.shape != gt.shape:
            raise ValueError(f"pred's shape {pred.shape} should be equal to gt's shape {gt.shape}")

        mask = (gt >= 0) & (gt < class_num)
        conf_matrix = np.bincount(
            class_num * gt[mask].astype(np.int64) + pred[mask].astype(np.int64),
            minlength=class_num ** 2,
        ).reshape(class_num, class_num)
        return conf_matrix

    def update(self, pred: np.ndarray, gt: np.ndarray):
        """Update confusion matrix accroding to input pred and gt

        Function get pred and gt array from input, which mean predicting label and ground truth label respectively.
        The label can be class of the input image, or class of each pixel of the input image. As long as pred and gt
        have the same shape.

        Function will use get_conf_matrix to calculate the last_conf_matrix of current input then add it to
        self.conf_matrix

        Args:
            pred: Numpy array of pred label
            gt: Numpy array of ground truth label.
        """
        super(ClassificationMetric, self).update(pred, gt)

        self._last_conf_matrix = self.get_conf_matrix(pred, gt, self.class_num)
        self.conf_matrix += self._last_conf_matrix

    def _plot_conf_matrix(self, **kwargs) -> plt.Figure:
        df = pd.DataFrame(self.conf_matrix, index=self.class_names, columns=self.class_names)

        kwargs["figsize"] = [self.class_num * 2, self.class_num * 2] if "figsize" not in kwargs else kwargs["figsize"]
        return pretty_plot_confusion_matrix(df, pred_val_axis='x', **kwargs)

    def show_conf_matrix(self, **kwargs) -> plt.Figure:
        """Plot confusion matrix

        Args:
            **kwargs: kwargs for plt.figure()

        Returns:
            Shown figure
        """
        fig = self._plot_conf_matrix(**kwargs)
        fig.show()
        plt.show()

        return fig

    def save_metric(self, save_dir: str='.', importance: int=0, **kwargs):
        """Saving Metric with confusion matrix and statistics

        Args:
            save_dir: Dictionary where metrics saved
            importance: Only statistics' statistic.importance > importance will be saved
            **kwargs: keywords arguments for plotting and saving confusion matrix plot confusion_matrix.png
        """
        # * Save statistics
        self.save_statistics(save_dir, importance)

        # * Save confusion matrix
        conf_matrix_txt = os.path.join(save_dir, 'confusion_matrix.txt')
        np.savetxt(conf_matrix_txt, self.conf_matrix, fmt="%d")

        # * Save confusion matrix plot
        try:
            fig = self._plot_conf_matrix(**kwargs)
            fig.savefig(os.path.join(save_dir, 'confusion_matrix.png'))
        except Exception:
            rprint(f"可视化混淆矩阵失败！")
            print(traceback.format_exc())

    @Statistic.getter(0)
    def samples_num(self):
        return self.conf_matrix.sum()

    N = samples_num

    @Statistic.getter(0)
    def condition_positive(self):
        return np.sum(self.conf_matrix, axis=1)

    @Statistic.getter(0)
    def condition_negative(self):
        return self.samples_num - self.condition_positive

    @Statistic.getter(0)
    def prediction_positive(self):
        return np.sum(self.conf_matrix, axis=0)

    @Statistic.getter(0)
    def prediction_negative(self):
        return self.samples_num - self.prediction_positive

    @Statistic.getter(0)
    def true_positive(self):
        return np.diag(self.conf_matrix)

    TP = true_positive

    @Statistic.getter(0)
    def false_negative(self):
        return self.condition_positive - self.true_positive

    FN = false_negative

    @Statistic.getter(0)
    def false_positive(self):
        return self.prediction_positive - self.true_positive

    FP = false_positive

    @Statistic.getter(0)
    def true_negative(self):
        return self.prediction_negative - self.false_negative

    TN = true_negative

    @Statistic.getter(1)
    def cls_weight(self):
        return self.condition_positive / self.N

    condition_weight = cls_weight

    @Statistic.getter(1)
    def prediction_weight(self):
        return self.prediction_positive / self.N

    @Statistic.getter(1)
    def cls_recall(self):
        return self.TP / self.condition_positive

    @Statistic.getter(1)
    def macro_avg_recall(self):
        return np.nanmean(self.cls_recall)

    @Statistic.getter(1)
    def weighted_avg_recall(self):
        return np.nansum(self.cls_weight * self.cls_recall)

    @Statistic.getter(1)
    def cls_precision(self):
        return self.TP / self.prediction_positive

    @Statistic.getter(1)
    def macro_avg_precision(self):
        return np.nanmean(self.cls_precision)

    @Statistic.getter(1)
    def weighted_avg_precision(self):
        return np.nansum(self.prediction_weight * self.cls_precision)

    @Statistic.getter(1)
    def cls_F1_score(self):
        return (2 * self.cls_precision * self.cls_recall) / (self.cls_precision + self.cls_recall)

    @Statistic.getter(2)
    def accuracy(self):
        return np.sum(self.TP) / self.N

    ACC = accuracy

    @Statistic.getter(2)
    def F1_score(self):
        return np.nanmean(self.cls_F1_score)

    F1 = F1_score

    @Statistic.getter(1)
    def weighted_F1_score(self):
        return np.nansum(self.cls_weight * self.cls_F1_score)


class SegmentationMetric(ClassificationMetric):

    @Statistic.getter(1)
    def cls_IoU(self):
        return self.TP / (self.FP + self.FN + self.TP)

    @Statistic.getter(1)
    def valid(self):
        return self.condition_positive > 0

    @Statistic.getter(2)
    def mIoU(self):
        return np.mean(self.cls_IoU[self.valid])

    @Statistic.getter(2)
    def weighted_mIoU(self):
        return np.sum(self.cls_weight[self.valid] * self.cls_IoU[self.valid])



# my_project/config.py
from yacs.config import CfgNode as CN


_C = CN()

_C.SYSTEM = CN()
# Number of GPUS to use in the experiment
_C.SYSTEM.NUM_GPUS = 8
# Number of workers for doing things
_C.SYSTEM.NUM_WORKERS = 4

_C.TRAIN = CN()
# A very important hyperparameter
_C.TRAIN.HYPERPARAMETER_1 = 0.1
# The all important scales for the stuff
_C.TRAIN.SCALES = (2, 4, 8, 16)


# def get_cfg_defaults():
#   """Get a yacs CfgNode object with default values for my_project."""
#   # Return a clone so that the defaults will not be altered
#   # This is for the "local variable" use pattern
#   return _C.clone()

# Alternatively, provide a way to import the defaults as
# a global singleton:
cfg = _C  # users can `from config import cfg`



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/5/28 20:37
@File    : cat_head.py
@Software: PyCharm
@Desc    : 
"""
from typing import Optional, Union, TYPE_CHECKING

from dataclasses import dataclass

from .config import Config

__all__ = ['喵', 'meow']


@dataclass
class CatHead(object):
    """该类型实例「喵」可作为全局变量使用。"""
    cfg: Optional['Config'] = None
    is_debug: bool = False
    rand_seed_final: Union[str, int, None] = None


喵 = meow = CatHead()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/6/22 22:28
@File    : test_param_lazy.py
@Software: PyCharm
@Desc    : 
"""
import pytest
import os.path as osp

import sys
sys.path = ['', 'py_tools/param_tuner/tests'] + sys.path  # noqa: E402

from alchemy_cat.py_tools import Cfg2Tune


@pytest.fixture(scope='function')
def param_lazy_config():
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'param_lazy_config', 'cfg.py'))


def test_param_lazy_config(param_lazy_config):
    def get_att_val(cfg):
        return [cfg.foo0.foo1.a, cfg.foo0.lazy0, cfg.lazy1, cfg.foo0.a_max]

    wanted_att_val = [
        [0, 10, -2, 2],
        [1, 11, -1, 2],
        [2, 12, 0, 2],
    ]

    att_val = [get_att_val(cfg) for cfg in param_lazy_config.get_cfgs()]
    assert wanted_att_val == att_val



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: __init__.py.py
@time: 2020/1/7 17:12
@desc:
"""
from .dataset import *
from .data_auger import *
from .utils import *
from .prefetcher import Prefetcher
from .data_manager import DataManager
from .inf_loader import inf_loader



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: my_process_bar.py
@time: 2019/7/30 2:29
@desc:
"""
import time
import numpy as np


class ProcessBar(object):

    def __init__(self, max_value, start_value=0, percent_step=0.01, lines_num=50):
        self.start_time = time.time()
        self.max_value = max_value
        self.start_value = start_value
        self.percent_step = percent_step
        self.lines_num = lines_num

        self.finished = False
        self.closed = False

        self.current_value = self.start_value
        self.current_percent = self.current_value / self.max_value
        self.current_time = self.start_time

        self.show_str = ''
        self._PrintBar()

    def _PrintBar(self):
        show_lines_num = int(np.round(self.lines_num * self.current_percent))
        lines = '|' * show_lines_num + ' ' * (self.lines_num - show_lines_num)
        self.show_str = "{percent:2d}%[{lines}]{delta_time:.2f}s".\
            format(**{'percent': int(np.round(self.current_percent * 100)), 'lines': lines,
                        'delta_time': time.time() - self.start_time})
        print(self.show_str, end='\r')

    def UpdateBar(self, current_value):
        if self.finished:
            return False
        if (current_value - self.current_value) / self.max_value < self.percent_step\
                and current_value != self.max_value:
            return False
        self.current_value = current_value
        self.current_percent = self.current_value / self.max_value
        self.current_time = time.time()
        self._PrintBar()
        if current_value == self.max_value:
            self.Close()
            self.finished = True
        return True

    def SkipMsg(self, msg: str, logger=None):
        print(' ' * (len(self.show_str) + 1), end='\r')
        if logger:
            logger.info(msg)
        else:
            print(msg)
        self._PrintBar()

    def Close(self, msg=None, logger=None):
        if not self.closed:
            print('')
            if isinstance(msg, str):
                if logger:
                    logger.info(msg)
                else:
                    print(msg)


if __name__ == "__main__":
    N = 1000
    process_bar = ProcessBar(N - 1)
    for i in range(N):
        process_bar.UpdateBar(i)
        time.sleep(0.001)
    print('Test')




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: normalize_tensor.py
@time: 2020/3/29 22:37
@desc:
"""
import numpy as np
import torch

__all__ = ['min_max_norm', 'channel_min_max_norm']


def min_max_norm(arr: torch.Tensor | np.ndarray, dim: int | tuple, detach_min_max: bool=True,
                 thresh: float | None =None, eps: float=1e-7) -> torch.Tensor:
    """在指定维度min-max归一化输入张量。

    Args:
        arr: 输入张量。
        dim: 做归一化的维度。
        detach_min_max: 是否分离min和max。
        thresh: 若不为None，将归一化前张量中小于thresh的值置为thresh。
        eps: 防止除0错误的极小值。

    Returns:
        归一化后的张量。
    """
    # * 若为numpy数组，转为torch张量。
    if is_numpy := isinstance(arr, np.ndarray):
        arr = torch.from_numpy(arr)

    # * 应用阈值。
    if thresh is not None:
        arr = torch.maximum(arr, torch.tensor(thresh, device=arr.device, dtype=arr.dtype))

    # * 归一化。
    arr_min, arr_max = arr.amin(dim=dim, keepdim=True), arr.amax(dim=dim, keepdim=True)
    if detach_min_max:
        arr_min, arr_max = arr_min.detach(), arr_max.detach()
    arr_normed = (arr - arr_min) / (arr_max - arr_min + eps)  # 其实用max(max - min, 1e-7)更安全（已经归一化过的不变）。

    # * 若为numpy数组，转为numpy数组。
    if is_numpy:
        arr_normed = arr_normed.numpy()

    return arr_normed


def channel_min_max_norm(arr: torch.Tensor, eps: float=1e-5) -> torch.Tensor:
    """Min-Max normalization for (N, C, ...) tensor

    Args:
        arr: (N, C, ...) tensor
        eps: eps added to divider

    Returns:
        Min-Max normalized tensor in dim behind C
    """
    batch_arr = arr.view(arr.shape[0], arr.shape[1], -1)  # (N, C, cumprod(...))

    batch_min, _ = batch_arr.min(dim=-1, keepdim=True)
    batch_arr = batch_arr - batch_min  # Don't modify origin arr

    batch_max, _ = batch_arr.max(dim=-1, keepdim=True)
    batch_arr /= (batch_max + eps)  # Inplace Operation

    return batch_arr.view_as(arr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/4/27 15:13
@File    : boundary_recall.py
@Software: PyCharm
@Desc    : 
"""
from typing import Optional

import numpy as np
from skimage.segmentation import find_boundaries
from skimage.morphology import disk, dilation, binary_dilation

from alchemy_cat.py_tools import Tracker, Statistic

__all__ = ["BoundaryRecall"]


class BoundaryRecall(Tracker):

    def __init__(self, ignore_label: Optional[int] = None, ignore_boundary_width: Optional[int] = None,
                 tolerance_radius: Optional[int] = 2):
        """Calculate Boundary Recall for superpixels

        Args:
            ignore_label: If boundary is marked with ignore label(like VOC non-aug), set ignore_label to eliminate
                ignore_label boundary.
            ignore_boundary_width: Boundary width of which marked by boundary.
            tolerance_radius: Max tolerance distance between superpixel boundary and gt boundary.
        """
        super().__init__()

        if ignore_label is not None:
            assert ignore_boundary_width is not None

        self.ignore_label = ignore_label
        self.ignore_boundary_width = ignore_boundary_width
        self.tolerance_radius = tolerance_radius

        self._fg_fg_recall_num = self._fg_bg_recall_num = 0
        self._fg_fg_gt_num = self._fg_bg_gt_num = 0

    def update(self, superpixel: np.ndarray, gt: np.ndarray):
        super().update(superpixel, gt)

        # * gt and superpixel should have same shape
        assert superpixel.shape == gt.shape

        # * Eliminate ignore label expressed boundary
        if self.ignore_label is not None:
            gt = self._eliminate_ignore_boundary_in_gt(gt)

        # * Get fg, bg mask
        fg_mask = gt > 0

        # * Get sp dilated boundary
        sp_boundary = find_boundaries(superpixel, connectivity=superpixel.ndim, mode='thick')
        dilated_sp_boundary = binary_dilation(sp_boundary, selem=disk(self.tolerance_radius)).astype(np.uint8)

        # * Get fg_fg_gt_boundary
        fg_fg_gt_boundary = find_boundaries(gt, connectivity=gt.ndim, mode='outer').astype(np.uint8)
        fg_fg_gt_boundary *= fg_mask

        # * Get fg_bg_gt_boundary
        fg_bg_gt_boundary = find_boundaries(gt, connectivity=gt.ndim, mode='thick').astype(np.uint8)
        fg_bg_gt_boundary -= fg_fg_gt_boundary

        # * update boundary recall
        self._fg_fg_recall_num += np.count_nonzero(dilated_sp_boundary * fg_fg_gt_boundary)
        self._fg_bg_recall_num += np.count_nonzero(dilated_sp_boundary * fg_bg_gt_boundary)

        # * update gt num
        self._fg_fg_gt_num += np.count_nonzero(fg_fg_gt_boundary)
        self._fg_bg_gt_num += np.count_nonzero(fg_bg_gt_boundary)

    def _eliminate_ignore_boundary_in_gt(self, gt_: np.ndarray):
        if np.any(gt_ == self.ignore_label):
            # * Set ignore to background
            gt = gt_.copy()
            gt[gt == self.ignore_label] = 0
            # * dilate with cross-shaped structuring element (connectivity=1) for ignore_boundary_width // 2 times.
            for _ in range(self.ignore_boundary_width // 2):
                gt = dilation(gt)
        else:
            gt = gt_
        assert not np.any(gt == self.ignore_label)
        return gt

    @Statistic.getter(importance=0)
    def fg_fg_boundary_recall(self):
        return self._fg_fg_recall_num / self._fg_fg_gt_num

    @Statistic.getter(importance=0)
    def fg_bg_boundary_recall(self):
        return self._fg_bg_recall_num / self._fg_bg_gt_num

    @Statistic.getter(importance=0)
    def boundary_recall(self):
        return (self._fg_fg_recall_num + self._fg_bg_recall_num) / (self._fg_fg_gt_num + self._fg_bg_gt_num)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/5/3 14:09
@File    : samp.py
@Software: PyCharm
@Desc    : 采样工具包。
"""
from typing import Union, Optional, List

import math

import torch

__all__ = ['shuffle_cat', 'samp_idx_with_1d_mask', 'samp_idxes_with_2d_mask', 'top_k_on_mask', 'samp_on_2d_mask']


def _get_rank_argsort_argsort(val: torch.Tensor, reverse: bool=False, dim: int=1) -> torch.Tensor:
    return val.argsort(dim=dim, descending=reverse).argsort(dim=dim)


def _get_rank_argsort_scatter(val: torch.Tensor, reverse: bool=False) -> torch.Tensor:
    idx_sorted = val.argsort(dim=1, descending=reverse)
    rank_to_scatter = torch.arange(val.shape[1], dtype=torch.int32, device=val.device)[None, :].expand(val.shape)
    rank = torch.zeros_like(rank_to_scatter)
    return rank.scatter_(1, idx_sorted, rank_to_scatter)


def _get_rank_with_split(val: torch.Tensor, reverse: bool=False, split_size: int=2000) -> torch.Tensor:
    rank = torch.empty_like(val, dtype=torch.int32)
    for i in range(math.ceil(val.shape[0] / split_size)):
        rank[split_size * i:min(split_size * (i + 1), rank.shape[0]), :] = \
            _get_rank_argsort_scatter(val[split_size * i:min(split_size * (i + 1), rank.shape[0]), :], reverse)
    return rank


def shuffle_cat(tensor: torch.Tensor, shuffled_len: int=None, shuffle_all: bool=False,
                g: Optional[torch.Generator]=None) -> torch.Tensor:
    """将张量沿着0维多次洗牌后拼接到shuffled_len长度。

    Args:
        tensor: 被洗牌、拼接的的(N, ...)张量。
        shuffled_len: 洗牌、拼接后的长度。若为None，则为tensor.shape[0]。
        shuffle_all: 是否要完全洗牌，即于洗牌、拼接后的张量上，再做一次整体洗牌。
        g: torch的随机序列生成器。注意若不为None，则要和tensor在同一设备上。

    Returns:
        洗牌、拼接后的(shuffled_len, ...)张量。
    """
    if shuffled_len is None:
        shuffled_len = tensor.shape[0]

    weights = torch.ones((math.ceil(shuffled_len / tensor.shape[0]), tensor.shape[0]),
                         dtype=torch.float32, device=tensor.device)
    shuffled_idxes = torch.multinomial(weights, tensor.shape[0], replacement=False, generator=g)
    shuffled_idxes = shuffled_idxes.view(-1)[:shuffled_len]

    if shuffle_all:
        shuffled_idxes = shuffle_cat(shuffled_idxes, shuffled_idxes.shape[0], shuffle_all=False, g=g)

    return tensor[shuffled_idxes]


def samp_idx_with_1d_mask(mask: torch.Tensor, samp_num: int, resamp_lim: int=1,
                          g: Optional[torch.Generator] = None) -> torch.Tensor:
    """在1维掩码为真处，尝试采样samp_num个索引。若True个数小于samp_num，则尝试多轮采样采足samp_num个索引。
    但每个索引被重采样的次数不得超过resamp_lim次。

    Args:
        mask: 1维布尔掩码。
        samp_num: 尝试采样多少个索引。
        resamp_lim: 每个索引至多比重采样几次。
        g: torch的随机序列生成器。注意若不为None，则要和mask在同一设备上。

    Returns:
        采样后的1维索引。
    """
    assert mask.ndim == 1
    # * 获取mask为True位置的索引，即可采索引。
    idx: torch.Tensor = torch.nonzero(mask, as_tuple=True)[0]
    if idx.shape[0] == 0:
        return idx
    # * 计算（最大）实采数。
    max_final_samp_num = idx.shape[0] * resamp_lim
    final_samp_num = min(samp_num, max_final_samp_num)
    # * 将可采索引洗牌拼接到洗牌拼接到实采数长度并返回。
    return shuffle_cat(idx, final_samp_num, shuffle_all=False, g=g)


def samp_idxes_with_2d_mask(mask: torch.Tensor, samp_nums: Union[int, torch.Tensor], resamp_lim: int=1,
                            g: Optional[torch.Generator] = None) -> List[torch.Tensor]:
    """对2维掩码的每一行i，做samp_idx_with_1d_mask(mask[i], samp_nums[i], ...)。返回List[每行采出索引]

    Args:
        mask: 2维布尔掩码。
        samp_nums: 尝试采样多少个索引。若为int，则每行尝试采样samp_nums个索引；若为Tensor，则行i尝试采样sump_nums[i]个索引。
        resamp_lim: 每个索引至多比重采样几次。
        g: torch的随机序列生成器。注意若不为None，则要和mask在同一设备上。

    Returns:
        List[各行的采出索引]。
    """
    assert mask.ndim == 2
    if torch.is_tensor(samp_nums):
        assert samp_nums.shape[0] == mask.shape[0]

    samped_idxes = []
    # * 对mask的每一行，做带mask索引采样。
    for r in range(mask.shape[0]):
        samp_num = samp_nums[r].item() if torch.is_tensor(samp_nums) else samp_nums
        samped_idxes.append(samp_idx_with_1d_mask(mask[r], samp_num, resamp_lim, g))
    return samped_idxes


def _rank_with_mask(val: torch.Tensor, mask: torch.Tensor, reverse: bool=False) -> torch.Tensor:
    """在mask前景位置，沿着dim，得到val的排名。mask背景位置的排名总是为-1。

    Args:
        val: 被排序的值。
        mask: 掩码，mask背景位置的值不参与排名。
        reverse: 若为True，则从大到小排序，反之从小大大排序。

    Returns:
        val同尺寸张量，在mask前景位置，表示val沿着dim维的排名，背景位置总是为-1。
    """
    assert mask.dtype == torch.bool
    assert torch.is_floating_point(val)
    assert val.shape == mask.shape
    assert val.device == mask.device
    # * 得到排序值。若不reverse（从小到大排），排序值的背景部分为inf，反之为-inf。
    sort_val = val.clone()
    sort_val[~mask] = float('inf') if not reverse else -float('inf')
    # * 根据排序值，得到排名。
    rank = _get_rank_with_split(sort_val, reverse=reverse)
    # * 背景部分排名设为-1，返回。
    rank[~mask] = -1
    return rank


def top_k_on_mask(val: torch.Tensor, mask: torch.Tensor, k: int, reverse: bool=False) -> torch.Tensor:
    """沿着指定维度，将val前景（由mask指定）的top-k设为True。若前景不足k个，则所有前景均为True。

    Args:
        val: 用来寻找top-k的值。
        mask: 掩码，指定val的前背景。
        k: top-k的k值。
        reverse: 若为True，寻找最大的top-k，反之寻找最小的top-k。

    Returns:
        val同尺寸张量，沿着dim维度，top-k前景为True，反之为False。
    """
    rank = _rank_with_mask(val, mask, reverse)
    return (rank >= 0) & (rank < k)


def samp_on_2d_mask(mask: torch.Tensor, samp_nums: Union[int, torch.Tensor],
                    g: Optional[torch.Generator] = None) -> torch.Tensor:
    """沿着2维掩码的第二维，尝试采样samp_nums个前景设为True。若前景数小于采样数，则保留所有前景。

    Args:
        mask: 2维掩码。
        samp_nums: 尝试采样多少个前景。若为int，则每行尝试采样samp_nums个索引；若为Tensor，则行i尝试采样sump_nums[i]个索引。
        g: torch的随机序列生成器。注意若不为None，则要和mask在同一设备上。

    Returns:
        mask同尺度张量，采出前景为True，其余为False。
    """
    assert mask.ndim == 2
    if torch.is_tensor(samp_nums):
        assert samp_nums.shape[0] == mask.shape[0]
    # * 得到mask每行元素得到一个不重复的随机值。
    weights = torch.ones_like(mask, dtype=torch.float32)
    rand_val = torch.multinomial(weights, mask.shape[1], replacement=False, generator=g).to(torch.float32)
    # * 根据随机值排序，mask的前景得到一个随机排名。
    rand_rank = _rank_with_mask(rand_val, mask)
    # * 返回随机排名小于采样数的前景。
    if torch.is_tensor(samp_nums):
        return (rand_rank >= 0) & (rand_rank < samp_nums[:, None])
    else:
        return (rand_rank >= 0) & (rand_rank < samp_nums)



# -- configs/mnist/plain_usage,cifar10/cfg.py --

from alchemy_cat.dl_config import Config

cfg = Config(caps=('configs/mnist/plain_usage/cfg.py', 'alchemy_cat/dl_config/examples/configs/patches/cifar10.py'))
